# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.5.1"
__date__ = "12/03/2025"

"""
Updates:
    09/05/2025, v0.6.0
        - Added function get_list_dir to scan subdirectories, returns a list
          of subdirectories.

    12/03/2025, v0.5.1
        - Adapted to biomdp with translations.

    05/03/2025, v0.5.0
        - Removed function create_time_series_xr
          (included in file create_time_series.py).

    14/01/2025, v0.4.1
        - Añadida función round_to_nearest_even_2_decimal

    16/12/2024, v0.4.0
        - Incluida función procesaEMG, que estaba en nexus_processing_functions.
    
    13/12/2024, v0.3.2
        - Corregido nanargmax_xr, no incluía bien la dimensión.
        - Cambiado nombre _cross_correl_rapida_aux por _cross_correl_noisy_aux
    
    17/11/2024, v0.3.1
        - Corregido nanargmax_xr, no incluía bien la dimensión
        - Incluido nanargmin_xr

    04/09/2024, v0.3.0
        - Versión de crosscorrelation simple con Polars, mucho más rápida    

    29/08/2024, v0.2.0
        - Añadidas funciones auxiliares para calcular cross correlation. 

    17/08/2024, v0.1.0
        - Incluidas algunas funciones generales. 

"""

import time
from typing import List

import numpy as np
import scipy.integrate as integrate
import xarray as xr

# =============================================================================
# %% Functins
# =============================================================================


def integrate_window(
    daData: xr.DataArray,
    daWindow: xr.DataArray | None = None,
    daOffset: xr.DataArray | None = None,
    result_return: str = "continuous",
) -> xr.DataArray:
    """
    Integrates time series data within a specified window, with an optional offset.

    Parameters
    ----------
    daData : xarray.DataArray
        The data to be integrated. It should have a 'time' dimension.
    daWindow : xarray.DataArray, optional
        The window specifying the start ('ini') and end ('fin') events for integration.
        If None, defaults to the full range of daData.
    daOffset : xarray.DataArray, optional
        Offset to be subtracted from daData before integration. If None, defaults to zero.
    result_return : str, optional
        Specifies the type of result to return.
        "continuous" : returns a continuous time series with integrated signal
        "discrete" : returns discrete integrated values at the end of the window.
        Default is "continuous".

    Returns
    -------
    xarray.DataArray
        The integrated data, either as a continuous signal or discrete values depending on
        the value of result_return.

    Notes
    -----
    This function is useful for computing the integral of a time series data over a specified
    window. It allows for offset correction before integration and supports both continuous
    and discrete result types.
    """

    # If empty, fill daWindow with first and last data
    if daWindow is None:
        daWindow = (
            xr.full_like(daData.isel(time=0).drop_vars("time"), np.nan).expand_dims(
                {"event": ["ini", "fin"]}, axis=-1
            )
        ).copy()
        daWindow.loc[dict(event=["ini", "fin"])] = np.array([0, len(daData.time)])

    if daOffset is None:
        daOffset = (xr.full_like(daData.isel(time=0).drop_vars("time"), 0.0)).copy()

    if result_return == "discrete":

        def _integrate_discrete(data, t, offset, ini, fin, ID):
            if np.isnan(ini) or np.isnan(fin):
                return np.nan
            ini = int(ini)
            fin = int(fin)
            # print(ID)
            # plt.plot(data[ini:fin])
            try:
                dat = integrate.cumulative_trapezoid(
                    data[ini:fin] - offset, t[ini:fin], initial=0
                )[-1]
            except Exception as e:
                if ID is None:
                    print(f"Error integrating in {ID}. {e}")
                else:
                    print(f"Error integrating. {e}")
                dat = np.nan
            return dat

        """
        data = daData[0,0].data
        t = daData.time.data
        ini = daWindow[0,0].isel(event=0).data
        fin = daWindow[0,0].isel(event=1).data
        offset = daOffset[0,0].data
        """
        daInt = xr.apply_ufunc(
            _integrate_discrete,
            daData,
            daData.time,
            daOffset,
            daWindow.isel(event=0),
            daWindow.isel(event=1),
            daData.ID,
            input_core_dims=[["time"], ["time"], [], [], [], []],
            # output_core_dims=[['time']],
            exclude_dims=set(("time",)),
            vectorize=True,
            # join='exact',
        )

    elif result_return == "continuous":

        def _integrate_continuous(data, time, peso, ini, fin):
            # if np.count_nonzero(~np.isnan(data))==0:
            #     return np.nan
            dat = np.full(len(data), np.nan)
            try:
                ini = int(ini)
                fin = int(fin)
                # plt.plot(data[ini:fin])
                dat[ini:fin] = integrate.cumulative_trapezoid(
                    data[ini:fin] - peso, time[ini:fin], initial=0
                )
                # plt.plot(dat)
            except Exception as e:
                print(f"Error integrating. {e}")
                pass  # dat = np.full(len(data), np.nan)
            return dat

        """
        data = daDatos[2,0].data #.sel(axis='z').data
        time = daDatos.time.data
        peso=daPeso[2,0].sel(stat='media').data
        ini = daEventos[2,0].sel(evento='iniMov').data
        fin = daEventos[2,0].sel(evento='finMov').data
        plt.plot(data[int(ini):int(fin)])
        """
        daInt = xr.apply_ufunc(
            _integrate_continuous,
            daData,
            daData.time,
            daOffset,
            daWindow.isel(event=0),
            daWindow.isel(event=1),
            input_core_dims=[["time"], ["time"], [], [], []],
            output_core_dims=[["time"]],
            # exclude_dims=set(('time',)),
            vectorize=True,
            join="exact",
        )

    else:
        raise ValueError("result_return must be 'continuous' or 'discrete'")

    daInt.attrs = daData.attrs

    return daInt


def detrend_dim(da: xr.DataArray, dim: str, deg: int = 1) -> xr.DataArray:
    """
    Detrend the signal along a single dimension.

    Parameters
    ----------
    da: xr.DataArray
        The data to be detrended.
    dim: str
        The dimension along which the detrending is performed.
    deg: int, optional
        The degree of the detrending polynomial. Default is 1 (linear).

    Returns
    -------
    daDetrended: xr.DataArray
        The detrended data.
    """

    p = da.polyfit(dim=dim, deg=deg)
    fit = xr.polyval(da[dim], p.polyfit_coefficients)

    return da - fit


def RMS(daData: xr.DataArray, daWindow: xr.DataArray | None = None) -> xr.DataArray:
    """
    Calculate RMS in dataarray with dataarray window
    """
    # If empty, fill daWindow with first and last data
    if daWindow is None:
        daWindow = (
            xr.full_like(daData.isel(time=0).drop_vars("time"), np.nan).expand_dims(
                {"event": ["ini", "fin"]}, axis=-1
            )
        ).copy()
        daWindow.loc[dict(event=["ini", "fin"])] = np.array([0, len(daData.time)])

    def _rms(data, ini, fin):
        if np.count_nonzero(~np.isnan(data)) == 0:
            return np.array(np.nan)
        data = data[int(ini) : int(fin)]
        data = data[~np.isnan(data)]
        return np.linalg.norm(data[~np.isnan(data)]) / np.sqrt(len(data))

    """
    data = daData[0,0,0].data
    ini = daWindow[0,0,0].sel(event='ini').data
    fin = daWindow[0,0,0].sel(event='fin').data
    """
    # daRecortado = recorta_ventana_analisis(daData, daWindow)
    daRMS = xr.apply_ufunc(
        _rms,
        daData,
        daWindow.isel(event=0),
        daWindow.isel(event=1),
        input_core_dims=[["time"], [], []],
        vectorize=True,
    )
    return daRMS


def calculate_distance(point1: xr.DataArray, point2: xr.DataArray):
    """
    Calculates the distance between two 3D points.
    Requires dimension with x, y, z coordinates with name 'axis'.
    """
    return np.sqrt(((point1 - point2) ** 2).sum("axis"))


# Función para detectar onsets
"""
Ref: Solnik, S., Rider, P., Steinweg, K., Devita, P., & Hortobágyi, T. (2010). Teager-Kaiser energy operator signal conditioning improves EMG onset detection. European Journal of Applied Physiology, 110(3), 489–498. https://doi.org/10.1007/s00421-010-1521-8

Función sacada de Duarte (https://nbviewer.org/github/BMClab/BMC/blob/master/notebooks/Electromyography.ipynb)
The Teager-Kaiser Energy operator to improve onset detection
The Teager-Kaiser Energy (TKE) operator has been proposed to increase the accuracy of the onset detection by improving the SNR of the EMG signal (Li et al., 2007).
"""


def tkeo(x: np.ndarray) -> np.ndarray:
    r"""Calculates the Teager-Kaiser Energy operator.

    Parameters
    ----------
    x : 1D array_like
        raw signal

    Returns
    -------
    y : 1D array_like
        signal processed by the Teager-Kaiser Energy operator

    Notes
    -----

    See this notebook [1]_.

    References
    ----------
    .. [1] https://github.com/demotu/BMC/blob/master/notebooks/Electromyography.ipynb

    """
    x = np.asarray(x)
    y = np.copy(x)
    # Teager-Kaiser Energy operator
    y[1:-1] = x[1:-1] * x[1:-1] - x[:-2] * x[2:]
    # correct the data in the extremities
    y[0], y[-1] = y[1], y[-2]

    return y


def process_EMG(
    daEMG: xr.DataArray,
    fr: float | None = None,
    fc_band: List[float] = [10, 400],
    fclow: float = 8,
    btkeo: bool = False,
) -> xr.DataArray:
    from biomdp.filter_butter import filter_butter, filter_butter_bandpass

    if fr is None:
        fr = daEMG.freq
    # Filtro band-pass
    daEMG_proces = filter_butter_bandpass(
        daEMG, fr=fr, fclow=fc_band[0], fchigh=fc_band[1]
    )
    # Centers signal, necessary?
    daEMG_proces = daEMG_proces - daEMG_proces.mean(dim="time")

    if btkeo:
        daEMG_proces = xr.apply_ufunc(
            tkeo,
            daEMG_proces,
            input_core_dims=[["time"]],
            output_core_dims=[["time"]],
            vectorize=True,
        )
    # Rectifica
    daEMG_proces = abs(daEMG_proces)
    # filtro low-pass
    daEMG_proces = filter_butter(daEMG_proces, fr=fr, fc=fclow, kind="low")

    # daEMG_proces.attrs['freq'] = daEMG.attrs['freq']
    # daEMG_proces.attrs['units'] = daEMG.attrs['units']
    # daEMG_proces.time.attrs['units'] = daEMG.time.attrs['units']
    daEMG_proces.attrs = daEMG.attrs
    daEMG_proces.name = "EMG"

    return daEMG_proces


# TODO: GENERALIZE THE NORMALIZING FUNCTION
def NormalizaBiela360_xr(
    daData, base_norm_horiz="time", graphs=False
):  # recibe da de daTodos. Versión con numpy
    if base_norm_horiz == "time":
        eje_x = daData.time
    elif base_norm_horiz in ["biela", "crank"]:
        try:
            eje_x = daData.sel(n_var="AngBiela", axis="y")
        except:
            eje_x = daData.sel(n_var="AngBiela")
    else:
        print("Normalizing base unknown")
        return

    def _normaliza_t_aux(
        data, x, base_norm_horiz
    ):  # Función auxiliar para normalizar con xarray
        # return tnorm(data, k=1, step=-361, show=False)[0]
        if np.isnan(data).all():
            data = np.full(361, np.nan)
        else:  # elimina los nan del final y se ajusta
            data = data[~np.isnan(data)]
            x = x[: len(data)]
            if base_norm_horiz in ["biela", "crank"]:
                x = np.unwrap(x)
                x = x - x[0]
            xi = np.linspace(0, x[-1], 361)
            data = np.interp(xi, x, data)  # tnorm(data, k=1, step=-361, show=False)[0]
        return data

    daNorm = xr.apply_ufunc(
        _normaliza_t_aux,
        daData,
        eje_x,
        base_norm_horiz,
        input_core_dims=[["time"], ["time"], []],
        output_core_dims=[["AngBielaInRepe"]],
        exclude_dims=set(("AngBielaInRepe",)),
        vectorize=True,
    ).assign_coords(
        dict(
            AngBielaInRepe=np.arange(
                361
            ),  # hay que meter esto a mano. Coords en grados
            AngBielaInRepe_rad=(
                "AngBielaInRepe",
                np.deg2rad(np.arange(361)),
            ),  # Coords en radianes
        )
    )
    daNorm.AngBielaInRepe.attrs["units"] = "deg"
    daNorm.AngBielaInRepe_rad.attrs["units"] = "rad"
    daNorm.name = daData.name
    daNorm.attrs["units"] = daData.attrs["units"]

    return daNorm


# from scipy import stats
def _cross_correl_simple_aux(datos1, datos2, ID=None):
    """
    Simple and slow but exact function for cross correlation.
    So far, data1 has to be the longest one.
    Uses polars, faster than numpy.

    Example:
    daCrosscorr = xr.apply_ufunc(
        _cross_correl_simple_aux,
        daInstrument1,
        daInstrument2,,

        input_core_dims=[
            ["time"],
            ["time"],
        ],
        output_core_dims=[["lag"]],
        exclude_dims=set(
            (
                "lag",
                "time",
            )
        ),
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
    ).dropna(dim="lag", how="all")
    """
    import polars as pl

    if ID is not None:
        print(ID)

    # pre-creates the array where to store the correlations of each offset
    corr = np.full(max(len(datos1), len(datos2)), np.nan)

    if np.isnan(datos1).all() or np.isnan(datos2).all():
        print(f"{ID} vacío")
        return corr

    try:
        # Remove nans from the end for function stats.pearson
        dat1 = datos1[~np.isnan(datos1)]
        dat2 = datos2[~np.isnan(datos2)]

        for i in range(0, dat1.size - dat2.size):
            # Versión Polars más rápida
            df = pl.from_numpy(
                np.vstack([dat1[i : i + dat2.size], dat2]),
                schema=["a", "b"],
                orient="col",
            )
            corr[i] = df.select(pl.corr("a", "b")).item()

            # scipy version slowler
            # corr[i] = stats.pearsonr(dat1[i : i + dat2.size], dat2).statistic
        # plt.plot(corr)

    except Exception as err:
        print("Computational error, possibly empty", err)

    return corr  # si hay algún error, lo devuelve vacío


def _cross_correl_noisy_aux(datos1, datos2, ID=None):
    """
    Fast but sometimes less accurate function for cross correlation.
    Good for noisy signals.
    """
    from scipy import signal

    # if ID is not None:
    #     print(ID)

    ccorr = np.full(max(len(datos1), len(datos2)), np.nan)

    if np.isnan(datos1).all() and np.isnan(datos2).all():
        return ccorr

    # Delete Nans
    dat1 = datos1[~np.isnan(datos1)]
    dat2 = datos2[~np.isnan(datos2)]

    # Normalize
    dat1 = (dat1 - np.mean(dat1)) / np.std(dat1)
    dat2 = (dat2 - np.mean(dat2)) / np.std(dat2)

    # Path with zeros
    if len(dat1) != len(dat2):
        if len(dat1) < len(dat2):
            dat1 = np.append(dat1, np.zeros(len(dat2) - len(dat1)))
        else:
            dat2 = np.append(dat2, np.zeros(len(dat1) - len(dat2)))

    # Compute cross-correlation
    c = signal.correlate(
        np.gradient(np.gradient(dat1)), np.gradient(np.gradient(dat2)), "full"
    )
    c = c[int(len(c) / 2) :]
    ccorr[: len(c)] = c
    # desfase = int(np.ceil(np.argmax(ccorr) - (len(ccorr)) / 2) + 1)

    return ccorr  # [int(len(ccorr) / 2) :]


def cross_correl_xr(
    da1: xr.DataArray, da2: xr.DataArray, func=_cross_correl_simple_aux
) -> xr.DataArray:
    """
    Apply the specified cross-correlation function to dataarrays
    """
    if "ID" in da1.coords:
        id = da1.ID
    else:
        id = None

    daCrosscorr = xr.apply_ufunc(
        func,
        da2,
        da1,
        id,
        input_core_dims=[
            ["time"],
            ["time"],
            [],
            # [],
            # [],
        ],
        output_core_dims=[["lag"]],  # datos que devuelve
        exclude_dims=set(
            (
                "lag",
                "time",
            )
        ),
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
        # kwargs=args_func_cortes,
    ).dropna(dim="lag", how="all")
    daCrosscorr = daCrosscorr.assign_coords(lag=range(len(daCrosscorr.lag)))
    return daCrosscorr


def _nanargmax(data: np.ndarray, ID) -> float:
    if np.count_nonzero(~np.isnan(data)) == 0:
        return np.array(np.nan)
    # if np.isnan(data).all():
    #     print('Error')
    #     return np.nan

    return float(np.nanargmax(data))


def nanargmax_xr(da: xr.DataArray, dim: str | None = None) -> xr.DataArray:
    """
    Replace with .idxmax() and .idxmin() ??
    data = da[0,0]
    """
    if dim is None:
        raise ValueError("dim must be specified")

    daResult = xr.apply_ufunc(
        _nanargmax,
        # daiSen.sel(articulacion='rodilla', lado='L', eje='x').dropna(dim='time'),
        da,
        da["ID"],
        input_core_dims=[
            [dim],
            [],
        ],
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
        # kwargs=args_func_cortes,
    )
    return daResult


def _nanargmin(data: np.array, ID) -> float:
    if np.count_nonzero(~np.isnan(data)) == 0:
        return np.array(np.nan)
    # if np.isnan(data).all():
    #     print('Error')
    #     return np.nan

    return float(np.nanargmin(data))


def nanargmin_xr(da: xr.DataArray, dim: str = None) -> xr.DataArray:
    """
    data = da[0,0]
    """
    if dim is None:
        raise ValueError("dim must be specified")

    daResult = xr.apply_ufunc(
        _nanargmin,
        # daiSen.sel(articulacion='rodilla', lado='L', eje='x').dropna(dim='time'),
        da,
        da["ID"],
        input_core_dims=[
            [dim],
            [],
        ],
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
        # kwargs=args_func_cortes,
    )
    return daResult


def round_to_nearest_even_2_decimal(number: float) -> float:
    """Rounds a float to the nearest even number with 2 decimal places"""
    rounded = np.round(number * 50) / 50  # Round to the nearest 0.02

    return rounded


# =============================================================================
# %% TESTS
# =============================================================================

if __name__ == "__main__":
    # =============================================================================
    # %%---- Create a sample
    # =============================================================================

    # from scipy.signal import butter, filtfilt
    from pathlib import Path

    import matplotlib.pyplot as plt
    import numpy as np

    # import pandas as pd
    import xarray as xr

    # import seaborn as sns
    from biomdp.create_time_series import create_time_series_xr

    rnd_seed = np.random.seed(
        12340
    )  # fija la aleatoriedad para asegurarse la reproducibilidad
    n = 10
    duration = 15
    freq = 200.0
    pre_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[25, 29],
            range_amp=[40, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    post_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[22, 26],
            range_amp=[36, 40],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_a = xr.concat([pre_a, post_a], dim="moment")
    var_a.sel(n_var="a").plot.line(x="time", col="ID", col_wrap=4)
    var_a.attrs["freq"] = freq

    # =============================================================================
    # %% TEST INTEGRATE
    # =============================================================================
    daWindow = (
        xr.full_like(var_a.isel(time=0).drop_vars("time"), np.nan).expand_dims(
            {"event": ["ini", "fin"]}, axis=-1
        )
    ).copy()
    daWindow.loc[dict(event=["ini", "fin"])] = np.array([100, -300])
    daWindow.loc[dict(event=["ini", "fin"], ID="00")] = np.array([0, len(var_a.time)])

    # Discrete
    integrate_window(var_a, result_return="discrete")
    integrate_window(var_a, daWindow, result_return="discrete")

    # Continuous
    integ = integrate_window(var_a, daWindow, result_return="continuous")
    integ.sel(n_var="a").plot.line(x="time", col="ID", col_wrap=4)

    integ = integrate_window(var_a, daWindow, daOffset=60, result_return="continuous")
    integ.sel(n_var="a").plot.line(x="time", col="ID", col_wrap=4)

    integ = integrate_window(
        var_a, daWindow, daOffset=var_a.mean("time"), result_return="continuous"
    )
    integ.sel(n_var="a").plot.line(x="time", col="ID", col_wrap=4)

    # =============================================================================
    # %% TEST RMS
    # =============================================================================
    daWindow = (
        xr.full_like(var_a.isel(time=0).drop_vars("time"), np.nan).expand_dims(
            {"event": ["ini", "fin"]}, axis=-1
        )
    ).copy()
    daWindow.loc[dict(event=["ini", "fin"])] = np.array([100, 300])
    daWindow.loc[dict(event=["ini", "fin"], ID="00")] = np.array([0, len(var_a.time)])

    RMS(var_a, daWindow)

    # =============================================================================
    # %% TEST CROSSCORREL
    # =============================================================================
    daInstrument1 = var_a.sel(n_var="a", moment="pre").drop_vars("moment")
    daInstrument2 = daInstrument1.isel(
        time=slice(int(4 * var_a.freq), int(8 * var_a.freq))
    )

    daCrosscorr = xr.apply_ufunc(
        _cross_correl_simple_aux,
        daInstrument1,
        daInstrument2,  # .sel(partID=da1['partID'], tiempo='pre'),
        # daInstrument1.time.size - daInstrument2.time.size,
        # daInstrument1['partID'],
        # daInstrument1['tiempo'],
        input_core_dims=[
            ["time"],
            ["time"],
            # [],
            # [],
            # [],
        ],
        output_core_dims=[["lag"]],
        exclude_dims=set(
            (
                "lag",
                "time",
            )
        ),
        # dataset_fill_value=np.nan,
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
        # kwargs=args_func_cortes,
    ).dropna(dim="lag", how="all")
    daCrosscorr = daCrosscorr.assign_coords(lag=range(len(daCrosscorr.lag)))
    daCrosscorr.plot.line(x="lag")
    nanargmax_xr(daCrosscorr, dim="lag")

    daCrosscorr_rap = xr.apply_ufunc(
        _cross_correl_noisy_aux,
        daInstrument1,
        daInstrument2,  # .sel(partID=da1['partID'], tiempo='pre'),
        input_core_dims=[
            ["time"],
            ["time"],
            # [],
        ],
        output_core_dims=[["lag"]],
        exclude_dims=set(
            (
                "lag",
                "time",
            )
        ),
        # dataset_fill_value=np.nan,
        vectorize=True,
        dask="parallelized",
        keep_attrs=False,
        # kwargs=args_func_cortes,
    ).dropna(dim="lag", how="all")
    daCrosscorr_rap = daCrosscorr_rap.assign_coords(lag=range(len(daCrosscorr_rap.lag)))
    daCrosscorr_rap.plot.line(x="lag")
    nanargmax_xr(daCrosscorr_rap, dim="lag")


def get_list_dir(n_path):
    """
    Recursively gets a list of all subdirectories in a given directory.
    Returns a list of Path objects.
    """
    from os import scandir
    from pathlib import Path

    subfolders = [f.path for f in scandir(n_path) if f.is_dir()]
    for n_path in list(subfolders):
        subfolders.extend(get_list_dir(n_path))
    subfolders = [Path(f) for f in subfolders]
    return subfolders



import unittest
import numpy as np
import xarray as xr
from biomdp.slice_time_series_phases import (
    detect_onset_detecta_aux,
    find_peaks_aux,
    detect_events,
    slice_time_series,
    trim_window,
    show_events,
)
from biomdp.create_time_series import create_time_series_xr


class TestSliceTimeSeriesPhases(unittest.TestCase):
    def setUp(self):
        # Setup code to create sample data
        self.rnd_seed = np.random.seed(12340)
        self.n = 5
        self.duration = 5
        self.freq = 200.0

        self.Pre_a = (
            create_time_series_xr(
                rnd_seed=self.rnd_seed,
                num_subj=self.n,
                fs=self.freq,
                IDini=0,
                range_offset=[25, 29],
                range_amp=[40, 45],
                range_freq=[1.48, 1.52],
                range_af=[0, 30],
                amplific_noise=[0.4, 0.7],
                fc_noise=[3.0, 3.5],
                range_duration=[self.duration, self.duration],
            )
            .expand_dims({"n_var": ["a"], "moment": ["pre"]})
            .transpose("ID", "moment", "n_var", "time")
        )
        self.Post_a = (
            create_time_series_xr(
                rnd_seed=self.rnd_seed,
                num_subj=self.n,
                fs=self.freq,
                IDini=0,
                range_offset=[22, 26],
                range_amp=[36, 40],
                range_freq=[1.48, 1.52],
                range_af=[0, 30],
                amplific_noise=[0.4, 0.7],
                fc_noise=[3.0, 3.5],
                range_duration=[self.duration, self.duration],
            )
            .expand_dims({"n_var": ["a"], "moment": ["post"]})
            .transpose("ID", "moment", "n_var", "time")
        )
        self.var_a = xr.concat([self.Pre_a, self.Post_a], dim="moment")

        self.daTotal = xr.concat([self.var_a], dim="n_var")
        self.daTotal.name = "Angle"
        self.daTotal.attrs["freq"] = 1 / (
            self.daTotal.time[1].values - self.daTotal.time[0].values
        )
        self.daTotal.attrs["units"] = "deg"
        self.daTotal.time.attrs["units"] = "s"

    def test_detect_events(self):
        daEvents = detect_events(
            data=self.daTotal, func_events=detect_onset_detecta_aux
        )
        self.assertIsInstance(daEvents, xr.DataArray)

    def test_slice_time_series(self):
        daEvents = detect_events(
            data=self.daTotal, func_events=detect_onset_detecta_aux
        )
        dacuts = slice_time_series(self.daTotal, daEvents)
        self.assertIsInstance(dacuts, xr.DataArray)

    def test_trim_window(self):
        daEvents = detect_events(
            data=self.daTotal, func_events=detect_onset_detecta_aux
        )
        window = xr.concat(
            [daEvents.min("n_event"), daEvents.max("n_event")], dim="n_event"
        )
        daTrimed = trim_window(self.daTotal, window)
        self.assertIsInstance(daTrimed, xr.DataArray)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import xarray as xr
import sys
from pathlib import Path

# Add the source directory to sys.path to import biomdp
sys.path.append(r"f:\Programacion\Python\Mios\biomdp\src")

try:
    from biomdp.fotogrammetry import (
        dlt_calib,
        dlt_recons_xr,
        dlt_calibrate,
        dlt_reconstruct,
    )
except ImportError as e:
    print(f"Error importing biomdp: {e}")
    sys.exit(1)


def test_multicam_dlt():
    print("Testing Multi-Camera DLT Reconstruction...")

    # 1. Define 3D Control Points (Cube)
    xyz = np.array(
        [
            [0, 0, 0],
            [0, 10, 0],
            [10, 10, 0],
            [10, 0, 0],
            [0, 0, 10],
            [0, 10, 10],
            [10, 10, 10],
            [10, 0, 10],
        ]
    )

    # 2. Define Synthetic Cameras (Viewpoints)
    # We will use dlt_calibrate to generate L parameters from known 2D projections
    # Let's project these 3D points onto 2D planes slightly different for each camera

    # Camera 1: Looking from front-left
    uv1 = np.array(
        [
            [500, 500],
            [500, 400],
            [600, 400],
            [600, 500],
            [520, 480],
            [520, 380],
            [620, 380],
            [620, 480],
        ]
    )

    # Camera 2: Looking from front-right
    uv2 = np.array(
        [
            [400, 500],
            [400, 400],
            [300, 400],
            [300, 500],
            [420, 480],
            [420, 380],
            [320, 380],
            [320, 480],
        ]
    )

    # Create Calibration Parameters
    nd = 3
    L1, err1 = dlt_calibrate(nd, xyz, uv1)
    L2, err2 = dlt_calibrate(nd, xyz, uv2)

    Ls = np.vstack([L1, L2])
    print(f"Calibration Errors: Cam1={err1:.4f}, Cam2={err2:.4f}")

    # 3. Prepare Data for dlt_recons_xr
    # We want to reconstruct the same points we used for calibration
    # Create DataArray with dimensions (time, camera, axis) or list of DataArrays

    # Using list of DataArrays (one per camera)
    n_points = xyz.shape[0]
    times = np.arange(n_points)

    da_uv1 = xr.DataArray(
        uv1,
        dims=("time", "axis"),
        coords={"time": times, "axis": ["x", "y"]},
        name="cam1",
    )

    da_uv2 = xr.DataArray(
        uv2,
        dims=("time", "axis"),
        coords={"time": times, "axis": ["x", "y"]},
        name="cam2",
    )

    # 4. Run Reconstruction
    try:
        # Pass list of DataArrays
        recons_da = dlt_recons_xr([da_uv1, da_uv2], Ls, num_dims=3, num_cams=2)

        print("\nReconstructed DataArray:")
        print(recons_da)

        # 5. Check Accuracy
        recons_xyz = recons_da.values

        # Calculate localized error
        diff = recons_xyz - xyz
        dist_err = np.linalg.norm(diff, axis=1)
        mean_err = np.mean(dist_err)

        print(f"\nMean Reconstruction Error: {mean_err:.6f}")

        if (
            mean_err < 1e-1
        ):  # Allow some tolerance due to the fake projection/calibration loop
            print("SUCCESS: Reconstruction is accurate.")
        else:
            print("FAILURE: Reconstruction error is too high.")

    except Exception as e:
        print(f"\nFAILURE with Exception: {e}")
        import traceback

        traceback.print_exc()


if __name__ == "__main__":
    test_multicam_dlt()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 09 11:12:37 2025

@author: Jose L. L. Elvira

Read data from Kistler Bioware .txt exported files.
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.2.1"
__date__ = "30/05/2025"

"""
Updates:
    30/05/2025, v0.2.1
        - Incuded "separator" parameter in read_kistler_txt_pl.

    25/03/2025, v0.2.0
        - Incuded general function to distribute according to "engine".
    
    09/03/2025, v0.1.0
        - Version imported from jump_forces_utils.

"""

# import polars as pl
import time
from pathlib import Path
from typing import Any, List

import numpy as np
import pandas as pd
import xarray as xr

# =============================================================================
# %% Functions
# =============================================================================


def read_kistler_txt(
    file: str | Path,
    lin_header: int = 17,
    n_vars_load: List[str] | None = None,
    # to_dataarray: bool = False,
    engine="polars",
    raw: bool = False,
    magnitude: str = "force",
) -> Any:
    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if engine == "polars":
        ret = read_kistler_txt_pl(
            file,
            lin_header=lin_header,
            n_vars_load=n_vars_load,
            raw=raw,
            magnitude=magnitude,
        )

    elif engine == "pandas":
        ret = read_kistler_txt_pd(
            file,
            lin_header=lin_header,
            n_vars_load=n_vars_load,
            raw=raw,
            magnitude=magnitude,
        )

    elif engine == "arrow":
        ret = read_kistler_txt_arrow(
            file,
            lin_header=lin_header,
            n_vars_load=n_vars_load,
            raw=raw,
            magnitude=magnitude,
        )
    else:
        raise ValueError(
            f"Engine {engine} not valid\nTry with 'polars', 'pandas' or 'arrow'"
        )

    return ret


# Carga un archivo de Bioware como dataframe de Polars
def read_kistler_txt_pl(
    file: str | Path,
    lin_header: int = 17,
    n_vars_load: List[str] | None = None,
    # to_dataarray: bool = False,
    raw: bool = False,
    magnitude: str = "force",
    separator: str = "\t",
) -> xr.DataArray | Any:
    """
    Read .txt files exported from Bioware

    Note: since BioWare Version 5.6.1.0 time column changed from "abs time (s)" to "abs time"
    """
    try:
        import polars as pl
    except:
        raise ImportError(
            "Polars package not instaled. Install it if you want to use the accelerated version"
        )

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    try:
        df = (
            pl.read_csv(
                file,
                has_header=True,
                skip_rows=lin_header,
                skip_rows_after_header=1,
                columns=n_vars_load,
                separator=separator,
            )  # , columns=nom_vars_cargar)
            # .slice(1, None) #quita la fila de unidades (N) #no hace falta con skip_rows_after_header=1
            # .select(pl.col(n_vars_load))
            # .rename({'abs time (s)':'time'}) #'Fx':'x', 'Fy':'y', 'Fz':'z',
            #          #'Fx_duplicated_0':'x_duplicated_0', 'Fy_duplicated_0':'y_duplicated_0', 'Fz_duplicated_0':'z'
            #          })
        ).with_columns(pl.all().cast(pl.Float64()))

        # ----Transform polars to xarray
        if raw:
            return df

        else:
            freq = 1 / (
                df[1, 0] - df[0, 0]  # assumes time in first column
            )  # (df[1, "abs time (s)"] - df[0, "abs time (s)"])
            if magnitude == "force":
                try:
                    x = df.select(pl.col("^*Fx.*$")).to_numpy()
                    y = df.select(pl.col("^*Fy.*$")).to_numpy()
                    z = df.select(pl.col("^*Fz.*$")).to_numpy()
                except:
                    raise Exception("Expected header with abs time (s), Fx, Fy, Fz")
                data = np.stack([x, y, z])

                # ending = -3
                coords = {
                    "axis": ["x", "y", "z"],
                    "time": np.arange(data.shape[1]) / freq,
                    "plate": range(
                        1, x.shape[1] + 1
                    ),  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                }
                da = (
                    xr.DataArray(
                        data=data,
                        dims=coords.keys(),
                        coords=coords,
                    )
                    .astype(float)
                    .transpose("plate", "axis", "time")
                )
                da.name = "Forces"
                da.attrs["freq"] = freq
                da.time.attrs["units"] = "s"
                da.attrs["units"] = "N"

            elif magnitude == "cop":
                try:
                    x = df.select(pl.col("^*Ax.*$")).to_numpy()
                    y = df.select(pl.col("^*Ay.*$")).to_numpy()
                except:
                    raise Exception("Expected header with abs time (s), Ax, Ay")
                data = np.stack([x, y])

                coords = {
                    "axis": ["x", "y"],
                    "time": np.arange(data.shape[1]) / freq,
                    "plate": range(
                        1, x.shape[1] + 1
                    ),  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                }
                da = (
                    xr.DataArray(
                        data=data,
                        dims=coords.keys(),
                        coords=coords,
                    )
                    .astype(float)
                    .transpose("plate", "axis", "time")
                )
                da.name = "COP"
                da.attrs["freq"] = freq
                da.time.attrs["units"] = "s"
                da.attrs["units"] = "mm"

            else:
                raise ValueError(
                    f"Magnitude {magnitude} not valid.\nTry with 'force' or 'cop'"
                )

            return da

    except Exception as err:
        print(f"\nATTENTION. Unable to read {file.name}, {err}, \n")


def read_kistler_txt_pd(
    file: str | Path,
    lin_header: int = 17,
    n_vars_load: List[str] | None = None,
    # to_dataarray: bool = False,
    raw: bool = False,
    magnitude: str = "force",
) -> xr.DataArray | pd.DataFrame | None:
    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    try:
        df = (
            pd.read_csv(
                file,
                header=lin_header,
                usecols=n_vars_load,  # ['Fx', 'Fy', 'Fz', 'Fx.1', 'Fy.1', 'Fz.1'], #n_vars_load,
                # skiprows=18,
                delimiter="\t",
                # dtype=np.float64,
                engine="c",  # "pyarrow" con pyarrow no funciona bien de momento cargar columnas con nombre repetido,
            ).drop(index=0)
        ).astype(float)
        # df.dtypes

        if raw:
            return df
        # ----Transform pandas to xarray
        else:
            try:
                freq = 1 / (df.loc[2, "abs time (s)"] - df.loc[1, "abs time (s)"])
            except:
                freq = 1 / (df.loc[2, "abs time"] - df.loc[1, "abs time"])
            if magnitude == "force":
                x = df.filter(regex="Fx")  # .to_numpy()
                y = df.filter(regex="Fy")
                z = df.filter(regex="Fz")
                data = np.stack([x, y, z])
                # ending = -3
                coords = {
                    "axis": ["x", "y", "z"],
                    "time": np.arange(data.shape[1]) / freq,
                    "plate": range(
                        1, x.shape[1] + 1
                    ),  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                }
                da = (
                    xr.DataArray(
                        data=data,
                        dims=coords.keys(),
                        coords=coords,
                    )
                    .astype(float)
                    .transpose("plate", "axis", "time")
                )
                # coords = {
                #     "axis": ["x", "y", "z"],
                #     "time": np.arange(data.shape[1]) / freq,
                #     "n_var": [
                #         "Force"
                #     ],  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                # }
                # da = (
                #     xr.DataArray(
                #         data=data,
                #         dims=coords.keys(),
                #         coords=coords,
                #     )
                #     .astype(float)
                #     .transpose("n_var", "axis", "time")
                # )
                da.name = "Forces"
                da.attrs["freq"] = freq
                da.time.attrs["units"] = "s"
                da.attrs["units"] = "N"

            elif magnitude == "cop":
                x = df.filter(regex="Ax")  # .to_numpy()
                y = df.filter(regex="Ay")
                data = np.stack([x, y])
                # ending = -3
                coords = {
                    "axis": ["x", "y"],
                    "time": np.arange(data.shape[1]) / freq,
                    "plate": range(
                        1, x.shape[1] + 1
                    ),  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                }
                da = (
                    xr.DataArray(
                        data=data,
                        dims=coords.keys(),
                        coords=coords,
                    )
                    .astype(float)
                    .transpose("plate", "axis", "time")
                )
                da.name = "COP"
                da.attrs["freq"] = freq
                da.time.attrs["units"] = "s"
                da.attrs["units"] = "mm"
            else:
                raise ValueError(
                    f"Magnitude {magnitude} not valid.\nTry with 'force' or 'cop'"
                )

            return da

    except Exception as err:
        print(f"\nATTENTION. Unable to read {file.name}, {err}, \n")


def read_kistler_txt_arrow(
    file: str | Path,
    lin_header: int = 17,
    n_vars_load: List[str] | None = None,
    # to_dataarray: bool = False,
    raw: bool = False,
    magnitude: str = "force",
) -> pd.DataFrame | xr.DataArray:
    """Testing, at the moment it does not work when there are repeated cols"""

    try:
        from pyarrow import csv
    except:
        raise ImportError("pyarrow not installed")

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    try:
        read_options = csv.ReadOptions(
            # column_names=['Fx', 'Fy', 'Fz'],
            skip_rows=lin_header,
            skip_rows_after_names=1,
        )
        parse_options = csv.ParseOptions(delimiter="\t")
        data = csv.read_csv(
            file, read_options=read_options, parse_options=parse_options
        )

        df = data.to_pandas()
        if raw:
            return df
        # ----Transform pandas to xarray
        else:
            if magnitude == "force":
                x = df.filter(regex="Fx")  # .to_numpy()
                y = df.filter(regex="Fy")
                z = df.filter(regex="Fz")
                data = np.stack([x, y, z])
                freq = 1 / (df.loc[2, "abs time (s)"] - df.loc[1, "abs time (s)"])
                ending = -3
                coords = {
                    "axis": ["x", "y", "z"],
                    "time": np.arange(data.shape[1]) / freq,
                    "plate": range(
                        1, x.shape[1] + 1
                    ),  # [x[:ending] for x in df.columns if 'x' in x[-1]],
                }
                da = (
                    xr.DataArray(
                        data=data,
                        dims=coords.keys(),
                        coords=coords,
                    )
                    .astype(float)
                    .transpose("plate", "axis", "time")
                )
                da.name = "Forces"
                da.attrs["freq"] = freq
                da.time.attrs["units"] = "s"
                da.attrs["units"] = "N"

            elif magnitude == "cop":
                raise Exception("Magnitude 'cop' not implemented yet")
            else:
                raise ValueError(
                    f"Magnitude {magnitude} not valid.\nTry with 'force' or 'cop'"
                )

            return da

    except Exception as err:
        print(f"\nATTENTION. Unable to read {file.name}, {err}, \n")


def split_plataforms(da: xr.DataArray) -> xr.DataArray:
    plat1 = da.sel(n_var=da.n_var.str.startswith("F1"))
    plat1 = plat1.assign_coords(n_var=plat1.n_var.str.lstrip("F1"))

    plat2 = da.sel(n_var=da.n_var.str.startswith("F2"))
    plat2 = plat2.assign_coords(n_var=plat2.n_var.str.lstrip("F2"))

    da = xr.concat([plat1, plat2], dim="plat").assign_coords(plat=[1, 2])

    return da


def split_axis(da: xr.DataArray) -> xr.DataArray:
    # NOT NECESSARY WITH COMPUTE_FORCES_AX???
    # TODO: The letter of the axis in the name must be removed
    x = da.sel(n_var=da.n_var.str.contains("x"))
    y = da.sel(n_var=da.n_var.str.contains("y"))
    z = da.sel(n_var=da.n_var.str.contains("z"))
    da = (
        xr.concat([x, y, z], dim="axis")
        # .assign_coords(n_var='plat1')
        .assign_coords(axis=["x", "y", "z"])
        # .expand_dims({'n_var':1})
    )
    return da


def compute_forces_axes(da: xr.DataArray) -> xr.DataArray:
    # da=daForce

    if "plat" not in da.coords:
        da = split_plataforms(da)

    Fx = da.sel(n_var=da.n_var.str.contains("x")).sum(dim="n_var")
    Fy = da.sel(n_var=da.n_var.str.contains("y")).sum(dim="n_var")
    Fz = da.sel(n_var=da.n_var.str.contains("z")).sum(dim="n_var")

    daReturn = xr.concat([Fx, Fy, Fz], dim="axis").assign_coords(axis=["x", "y", "z"])
    # daReturn.plot.line(x='time', col='plat')

    return daReturn


def compute_moments_axes(da: xr.DataArray) -> xr.DataArray:
    # da=daForce
    raise Exception("Not implemented yet")
    """
    if 'plat' not in da.coords:
        da = split_plataforms(da)

    Fx = da.sel(n_var=da.n_var.str.contains('x')).sum(dim='n_var')
    Fy = da.sel(n_var=da.n_var.str.contains('y')).sum(dim='n_var')
    Fz = da.sel(n_var=da.n_var.str.contains('z')).sum(dim='n_var')
        
    daReturn = (xr.concat([Fx, Fy, Fz], dim='axis')
                .assign_coords(axis=['x', 'y', 'z'])
                )
    #daReturn.plot.line(x='time', col='plat')
    """
    return daReturn


# =============================================================================
# %% TESTS
# =============================================================================
if __name__ == "__main__":
    # from biomdp.io.read_kistler_txt import read_kistler_c3d_xr, read_kistler_ezc3d_xr

    work_path = Path(r"src\datasets")
    file = work_path / "kistler_CMJ_1plate.txt"
    daForce = read_kistler_txt(file)
    daForce.isel(plate=0).plot.line(x="time")  # , col="plat")

    dfForce = read_kistler_txt(file, raw=True, engine="pandas")
    dfForce.plot(x="abs time (s)")

    dfForce = read_kistler_txt(file, raw=True, engine="arrow")
    dfForce.plot(x="abs time (s)")

    file = work_path / "kistler_DJ_2plates.txt"
    daForce = read_kistler_txt(file)
    daForce.plot.line(x="time", col="plate")
    daForce.sum(dim="plate").plot.line(x="time")

    daForce = read_kistler_txt(file, engine="pandas")
    daForce.plot.line(x="time", col="plate")

    daForce = read_kistler_txt(file, engine="arrow")
    daForce.plot.line(x="time", col="plate")

    dfForce = read_kistler_txt(file, raw=True, engine="pandas")
    dfForce[["Fx", "Fy", "Fz"]].plot()

    file = work_path / "kistler_COP.txt"
    daForce = read_kistler_txt(file, magnitude="cop")
    daForce.to_dataset("axis").plot.scatter(x="x", y="y", col="plate")

    daForce = read_kistler_txt(file, magnitude="cop", engine="pandas")
    daForce.to_dataset("axis").plot.scatter(x="x", y="y", col="plate")

    import timeit

    def test_performance():
        result = read_kistler_txt(file, engine="polars")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=50):.4f} s"
    )

    def test_performance():
        result = read_kistler_txt(file, engine="pandas")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=50):.4f} s"
    )

    def test_performance():
        result = read_kistler_txt(file, engine="arrow")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=50):.4f} s"
    )



# -*- coding: utf-8 -*-
"""
Created on Fry Sep 15 16:36:37 2023

@author: josel
"""

# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.3.1"
__date__ = "13/12/2025"

"""
Updates:
    13/12/2025, v0.3.1
        - Changed default engine to c3d.

    18/08/2025, v0.3.0
        - Added engine moveck_bridge_btk as 'mbbtk' (in development).

    26/04/2025, v0.2.1
            - Added function read_vicon_c3d_c3d_pose2sim (to explore).

    25/03/2025, v0.2.0
        - Incuded general function to distribute according to "engine".
    
    19/03/2025, v0.1.1
        - Adapted to biomdp with translations. 

    10/12/2024, v0.1.0
            - Comprueba si está instalado c3d y ezc3d, si no, avisa con mensaje.
            - Incluidas funciones con ezc3d.

    23/12/2023, v0.0.2
            - Perfeccionada carga de fuerzas.

    15/09/2023, v0.0.1
            - Empezado tomando trozos sueltos.
            
"""


import time
import warnings
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import xarray as xr


# =============================================================================
# %% Functions
# =============================================================================
def read_vicon_c3d(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
    engine: str = "c3d",
) -> xr.DataArray:
    if isinstance(file, str):
        file = Path(file)
    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # Ensures that file is a c3D file
    file = file.with_suffix(".c3d")

    if engine == "c3d":
        da = read_vicon_c3d_c3d(file, section, n_vars_load, coincidence)

    elif engine == "ezc3d":
        da = read_vicon_ezc3d(file, section, n_vars_load, coincidence)

    elif engine == "mbbtk":
        da = read_vicon_mbbtk(file, section, n_vars_load, coincidence)

    else:
        raise Exception(
            f"Engine {engine} not implemented.\nTry 'c3d', 'ezc3d' or 'mbbtk'"
        )

    return da


def read_vicon_c3d_c3d_pose2sim(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
) -> xr.DataArray:
    """tests with c3d read from pose2sim"""
    try:
        import c3d
    except:
        raise ImportError("Module c3d not installed.\nInstall with pip install c3d")

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # file =Path(r"F:\Programacion\Python\Mios\biomdp\src\datasets\vicon_CMJ_kinem_kinet_emg.c3d")
    # file =Path(r"F:\Programacion\Python\Mios\biomdp\src\datasets\vicon_CMJ_kinem_kinet_emg.c3d")

    # c3d header
    reader = c3d.Reader(open(file, "rb"))
    items_header = str(reader.header).split("\n")
    items_header_list = [item.strip().split(": ") for item in items_header]
    label_item = [item[0] for item in items_header_list]
    value_item = [item[1] for item in items_header_list]
    header_c3d = dict(zip(label_item, value_item))

    # unit
    for k1 in reader.group_items():
        if k1[0] == "POINT":
            for k2 in k1[1].param_items():
                if k2[0] == "UNITS":
                    if "mm" in k2[1].bytes[:].decode("utf-8"):
                        unit = "mm"
                        unit_scale = 0.001
                    else:
                        unit = "m"
                        unit_scale = 1  # mm

    # c3d data: reads 3D points (no analog data) and takes off computed data
    labels = reader.point_labels
    index_labels_markers = [
        i
        for i, s in enumerate(labels)
        if "Angle" not in s
        and "Power" not in s
        and "Force" not in s
        and "Moment" not in s
        and "GRF" not in s
    ]
    labels_markers = [labels[ind] for ind in index_labels_markers]

    index_data_markers = np.sort(
        np.concatenate(
            [
                np.array(index_labels_markers) * 3,
                np.array(index_labels_markers) * 3 + 1,
                np.array(index_labels_markers) * 3 + 2,
            ]
        )
    )
    t0 = int(float(header_c3d["first_frame"])) / int(float(header_c3d["frame_rate"]))
    tf = int(float(header_c3d["last_frame"])) / int(float(header_c3d["frame_rate"]))
    trc_time = np.linspace(
        t0, tf, num=(int(header_c3d["last_frame"]) - int(header_c3d["first_frame"]) + 1)
    )

    data = []
    for n, (i, points, _) in enumerate(list(reader.read_frames())):
        c3d_line = np.concatenate([item[:3] for item in points]) * unit_scale
        # c3d_line_markers = c3d_line[index_data_markers]
        data.append(points[:, :3]) * unit_scale
        # trc_line = '{i}\t{t}\t'.format(i=i, t=trc_time[n]) + '\t'.join(map(str,c3d_line_markers))
        # print(trc_line+'\n')
    data = np.concatenate(data)

    return data


def read_vicon_c3d_c3d(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
) -> xr.DataArray:
    """
    In c3d version 0.5.2, to avoid in16 error modify c3d.py file:
        Uncomment line 1530 (end_frame = param.uint32_value)
        Comment lines 1531 y 1532 (words = param.uint16_array
                                   end_frame = words[0] + words[1] * 65536
                                   )
    """

    try:
        import c3d
    except:
        raise ImportError("Module c3d not installed.\nInstall with pip install c3d")

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if section not in [
        "Trajectories",
        "Model Outputs",
        "Forces",
        "EMG",
    ]:
        raise Exception(
            'Section header not found, try "Trajectories", "Model Outputs", "Forces" or "EMG"'
        )
        return

    # timer = time.perf_counter()  # inicia el contador de tiempo

    try:
        # timerSub = time.perf_counter()  # inicia el contador de tiempo
        # print(f'Loading section {section}, file: {file.name}')

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with open(file, "rb") as handle:
                reader = c3d.Reader(handle)

                freq = reader.point_rate
                freq_analog = reader.analog_rate

                points = []
                analog = []
                for i, (_, p, a) in enumerate(reader.read_frames()):
                    points.append(p)
                    analog.append(a)
                    # if not i % 10000 and i:
                    #     print("Extracted %d point frames", len(points))

        # Trajectiories and Modeled outputs
        if section in ["Trajectories", "Model Outputs"]:
            labels = [s.replace(" ", "") for s in reader.point_labels]
            data = np.asarray(points)[:, :, :3]

            coords = {
                "time": np.arange(data.shape[0]) / freq,
                "n_var": labels,
                "axis": ["x", "y", "z"],
            }
            da = xr.DataArray(
                data,  # =np.expand_dims(data, axis=0),
                dims=coords.keys(),
                coords=coords,
                name="Trajectories",
                attrs={
                    "freq": freq,
                    "units": "mm",
                },
            ).transpose("n_var", "axis", "time")
            if "Trajectories" in section:
                # Delete unnamed trajectories and modeled outputs
                da = da.sel(
                    n_var=(
                        ~da.n_var.str.startswith("*") & ~da.n_var.str.contains("USERMO")
                    )
                )
            if "Model Outputs" in section:
                da = da.sel(n_var=da.n_var.str.contains("USERMO"))

        # Analogs
        elif section in ["Forces", "EMG"]:
            labels_analog = [
                s.split(".")[0].replace(" ", "") for s in reader.analog_labels
            ]
            data_analog = np.concatenate(analog, axis=1)

            # data_analog.shape
            coords = {
                "n_var": labels_analog,
                "time": np.arange(data_analog.shape[1]) / freq_analog,
            }
            da_analog = xr.DataArray(
                data=data_analog,
                dims=coords.keys(),
                coords=coords,
                attrs={"freq": freq_analog},
            )

            # Forces
            # Sometimes contains 'Force' and others 'Fx', Fy', 'Fz'
            # Get only force variables
            if section == "Forces":
                if "Force" in da_analog.n_var:  # newer versions of Nexus?
                    da = da_analog.sel(n_var=da_analog.n_var.str.contains("Force"))
                elif da_analog.n_var.str.contains(
                    "Fz"
                ).any():  # older versions of Nexus?
                    da = da_analog.sel(n_var=da_analog.n_var.str.startswith("F"))
                else:
                    da = xr.DataArray()
                    raise Exception("Apparently no force data in file")
                # if da_analog.n_var.str.startswith('F').any():#da_analog.n_var.str.contains('Force').any(): #'Force' in da_analog.n_var:
                try:
                    # da = da_analog.sel(n_var=da_analog.n_var.str.startswith('F')) #'Force') #.sel(n_var=da_analog.n_var.str.contains('Force'))
                    if len(da.n_var) == 3:  # 1 platform
                        x = da.isel(n_var=0)
                        y = da.isel(n_var=1)
                        z = da.isel(n_var=2)
                        da = (
                            xr.concat([x, y, z], dim="axis")
                            .assign_coords(n_var="plat1")
                            .assign_coords(axis=["x", "y", "z"])
                            .expand_dims({"n_var": 1})
                        )
                    elif len(da.n_var) == 6:  # 2 platforms
                        x = da.isel(n_var=[0, 3]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        y = da.isel(n_var=[1, 4]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        z = da.isel(n_var=[2, 5]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        da = (
                            xr.concat([x, y, z], dim="axis")
                            # .assign_coords(n_var=['plat1', 'plat2'])
                            .assign_coords(axis=["x", "y", "z"])
                        )
                    else:
                        raise Exception("The number of Force variables is not 3 or 6")
                    da.attrs["units"] = "N"
                    # da.time.attrs['units']='s'
                    # da.plot.line(x='time', col='axis', hue='n_var')
                except:
                    da = xr.DataArray()
                    raise Exception("Not available force data in file")

            # EMG
            elif section == "EMG":
                if da_analog.n_var.str.contains("EMG").any():
                    da = da_analog.sel(n_var=da_analog.n_var.str.contains("EMG"))
                    da.attrs["units"] = "mV"
                    # da.n_var.sortby('n_var')
                    # da.plot.line(x='time', col='n_var', col_wrap=3)
                else:
                    da = xr.DataArray()
                    raise Exception("No EMG data in file")

        da.time.attrs["units"] = "s"
        da.name = section

        # print('Time {0:.3f} s \n'.format(time.perf_counter()-timerSub))

    except Exception as err:
        da = xr.DataArray()
        print(f"\nATTENTION. Unable to process {file.name}, {err}\n")

    if n_vars_load is not None and "n_var" in da.coords:
        da = da.sel(n_var=n_vars_load)

    return da  # daTrajec, daModels, daForce, daEMG


def read_vicon_ezc3d(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
) -> xr.DataArray:
    try:
        import ezc3d
    except:
        raise ImportError(
            "Module ezc3d not installed.\nInstall with pip install ezc3d or conda install -c conda-forge ezc3d"
        )

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if section not in [
        "Trajectories",
        "Model Outputs",
        "Forces",
        "EMG",
    ]:
        raise Exception(
            'Section header not found, try "Trajectories", "Model Outputs", "Forces" or "EMG"'
        )
        return

    try:
        # timerSub = time.perf_counter()  # inicia el contador de tiempo
        # print(f'Loading section {section}, file: {file.name}')

        acq = ezc3d.c3d(file.as_posix())

        # Trajectiories and Modeled outputs
        if section in ["Trajectories", "Model Outputs"]:
            freq = acq["parameters"]["POINT"]["RATE"]["value"][0]

            labels = acq["parameters"]["POINT"]["LABELS"]["value"]
            data = acq["data"]["points"][:3, :, :]

            coords = {
                "axis": ["x", "y", "z"],
                "n_var": labels,
                "time": np.arange(data.shape[-1]) / freq,
            }
            da = (
                xr.DataArray(
                    data,  # =np.expand_dims(data, axis=0),
                    dims=coords.keys(),
                    coords=coords,
                    name="Trajectories",
                    attrs={
                        "freq": freq,
                        "units": "mm",
                    },
                )
                .transpose("n_var", "axis", "time")
                .dropna("time", how="all")  # para quitar los ceros cuano no hay dato
            )

            if "Trajectories" in section:
                # Delete unnamed trajectories and modeled outputs
                da = da.sel(
                    n_var=(
                        ~da.n_var.str.startswith("*") & ~da.n_var.str.contains("USERMO")
                    )
                )
            elif "Model Outputs" in section:
                da = da.sel(n_var=da.n_var.str.contains("USERMO"))

        # Analogs
        elif section in ["Forces", "EMG"]:
            # acq2 = ezc3d.c3d(file.as_posix(), extract_forceplat_data=True)
            # acq2["data"]["platform"][0]['center_of_pressure']

            freq = acq["parameters"]["ANALOG"]["RATE"]["value"][0]

            labels = acq["parameters"]["ANALOG"]["LABELS"]["value"]
            # acq["parameters"]["ANALOG"]['UNITS'].keys()

            data = acq["data"]["analogs"][0]

            # data_analog.shape
            coords = {
                "n_var": labels,
                "time": np.arange(data.shape[-1]) / freq,
            }
            da = xr.DataArray(
                data=data,
                dims=coords.keys(),
                coords=coords,
                attrs={"freq": freq},
            )

            # Forces
            # Sometimes contains 'Force' and others 'Fx', Fy', 'Fz'
            # Get only force variables
            if section == "Forces":
                if da.n_var.str.contains("Force").any():  # new versions of Nexus?
                    da = da.sel(n_var=da.n_var.str.contains("Force"))
                elif da.n_var.str.contains("Fz").any():  # old versions of Nexus?
                    da = da.sel(n_var=da.n_var.str.startswith("F"))
                else:
                    da = xr.DataArray()
                    raise Exception("Apparently no force data in file")
                # if da.n_var.str.startswith('F').any():#da.n_var.str.contains('Force').any(): #'Force' in da.n_var:
                try:
                    # da = da.sel(n_var=da.n_var.str.startswith('F')) #'Force') #.sel(n_var=da_analog.n_var.str.contains('Force'))
                    if len(da.n_var) == 3:  # 1 platform
                        x = da.isel(n_var=0)
                        y = da.isel(n_var=1)
                        z = da.isel(n_var=2)
                        da = (
                            xr.concat([x, y, z], dim="axis")
                            .assign_coords(n_var="plat1")
                            .assign_coords(axis=["x", "y", "z"])
                            .expand_dims({"n_var": 1})
                        )
                    elif len(da.n_var) == 6:  # 2 platforms
                        x = da.isel(n_var=[0, 3]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        y = da.isel(n_var=[1, 4]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        z = da.isel(n_var=[2, 5]).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        da = (
                            xr.concat([x, y, z], dim="axis")
                            # .assign_coords(n_var=['plat1', 'plat2'])
                            .assign_coords(axis=["x", "y", "z"])
                        )
                    elif len(da.n_var) == 8:  # Bioware 1 plat
                        da.attrs["description"] = "Data from Bioware 1 plat"
                    else:
                        raise Exception("The number of Force variables is not 3 or 6")

                    da.attrs["units"] = "N"
                    # da.time.attrs['units']='s'
                    # da.plot.line(x='time', col='axis', hue='n_var')

                except Exception as err:
                    da = xr.DataArray()
                    raise Exception(f"Not available force data in file, {err}")

            # EMG
            elif section == "EMG":
                if da.n_var.str.contains("EMG").any():
                    da = da.sel(n_var=da.n_var.str.contains("EMG"))
                    da.attrs["units"] = "V"
                    # da.n_var.sortby('n_var')
                    # da.plot.line(x='time', col='n_var', col_wrap=3)
                else:
                    da = xr.DataArray()
                    raise Exception("No EMG data found in file")

        da.time.attrs["units"] = "s"
        da.name = section

        # print('Tiempo {0:.3f} s \n'.format(time.perf_counter()-timerSub))

    except Exception as err:
        da = xr.DataArray()
        raise Exception(f"\nATTENTION. Unable to process {file.name}, {err}\n")

    if n_vars_load is not None and "n_var" in da.coords:
        da = da.sel(n_var=n_vars_load)

    return da  # daTrajec, daModels, daForce, daEMG


def read_vicon_mbbtk(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
) -> xr.DataArray:
    """
    IN DEVELOPMENT
    Reads a Vicon C3D file using the moveck_bridge_btk library.

    Args:
        file (str or Path): Path to the C3D file.
        section (str, optional): Section of data to load. Defaults to None.
        n_vars_load (List[str], optional): List of variables to load. Defaults to None.
        coincidence (str, optional): Method to match variables names. Defaults to "similar".

    Returns:
        xr.DataArray: A DataArray containing the loaded data.

    Raises:
        ImportError: If the moveck_bridge_btk module is not installed.

    """

    try:
        import moveck_bridge_btk as btk
    except:
        raise ImportError(
            "Module moveck_bridge_btk not installed.\nInstall with pip install moveck_bridge_btk"
        )

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if section not in [
        "Trajectories",
        "Model Outputs",
        "Forces",
        "EMG",
    ]:
        raise Exception(
            'Section header not found, try "Trajectories", "Model Outputs", "Forces" or "EMG"'
        )
        return

    try:
        # timerSub = time.perf_counter()  # inicia el contador de tiempo
        # print(f'Loading section {section}, file: {file.name}')

        h = btk.btkReadAcquisition(file.as_posix())

        # Trajectiories and Modeled outputs
        if section == "Trajectories":
            [markers, markersInfo] = btk.btkGetMarkers(h)
            freq = markersInfo["frequency"]

            labels = [n for n in markers if "*" not in n]
            data = []
            data.append([markers[s] for s in labels])
            data = np.concatenate(data)

            coords = {
                "n_var": labels,
                "time": np.arange(data.shape[1]) / freq,
                "axis": ["x", "y", "z"],
            }
            da = (
                xr.DataArray(
                    data,  # =np.expand_dims(data, axis=0),
                    dims=coords.keys(),
                    coords=coords,
                    name="Trajectories",
                    attrs={
                        "freq": freq,
                        "units": markersInfo["units"]["ALLMARKERS"],
                    },
                )
                .transpose("n_var", "axis", "time")
                .dropna("time", how="all")  # para quitar los ceros cuando no hay dato
            )

        # Model outputs
        elif section == "Model Outputs":
            raise Exception("Not implemented")
            """
            # Reads the same as markers???
            [points, pointsInfo] = btk.btkGetPoints(h)
            freq = btk.btkGetPointFrequency(h)
            
            labels = [n for n in points if '*' not in n]
            data=[]
            data.append([points[s] for s in labels])
            data = np.concatenate(data)
            
            coords = {
                "n_var": labels,
                "time": np.arange(data.shape[1]) / freq,
                "axis": ["x", "y", "z"],                
            }
            da = (
                xr.DataArray(
                    data,  # =np.expand_dims(data, axis=0),
                    dims=coords.keys(),
                    coords=coords,
                    name="Trajectories",
                    attrs={
                        "freq": freq,
                        "units": pointsInfo["units"]['ALL'],
                    },
                )
                .transpose("n_var", "axis", "time")
                .dropna("time", how="all")  # para quitar los ceros cuando no hay dato
            )
            """

        # Analogs
        elif section in ["Forces", "EMG", "EMGpl"]:
            [analogs_values, analog_infos] = btk.btkGetAnalogs(h)
            freq = btk.btkGetAnalogFrequency(h)

            # Forces
            if section == "Forces":
                labels = [
                    n
                    for n in analogs_values
                    if "Force." in n or n.startswith("F") and "Foot" not in n
                ]
                units = analog_infos["units"][labels[0]]
                data = []
                data.append([analogs_values[s] for s in labels])
                data = np.concatenate(data)

                coords = {
                    "n_var": labels,
                    "time": np.arange(data.shape[-1]) / freq,
                }
                da = xr.DataArray(
                    data=data,
                    dims=coords.keys(),
                    coords=coords,
                )

                try:
                    # da = da.sel(n_var=da.n_var.str.startswith('F')) #'Force') #.sel(n_var=da_analog.n_var.str.contains('Force'))
                    if len(da.n_var) == 3:  # 1 platform
                        x = da.isel(n_var=0)
                        y = da.isel(n_var=1)
                        z = da.isel(n_var=2)
                        da = (
                            xr.concat([x, y, z], dim="axis")
                            .assign_coords(n_var="plat1")
                            .assign_coords(axis=["x", "y", "z"])
                            .expand_dims({"n_var": 1})
                        )
                    elif len(da.n_var) == 6:  # 2 platforms
                        x = da.sel(n_var=da.n_var.str.contains("Fx")).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        y = da.sel(n_var=da.n_var.str.contains("Fy")).assign_coords(
                            n_var=["plat1", "plat2"]
                        )
                        z = da.sel(n_var=da.n_var.str.contains("Fz")).assign_coords(
                            n_var=["plat1", "plat2"]
                        )

                        da = (
                            xr.concat([x, y, z], dim="axis")
                            # .assign_coords(n_var=['plat1', 'plat2'])
                            .assign_coords(axis=["x", "y", "z"])
                        )
                        da.attrs["units"] = "N"
                    elif len(da.n_var) == 8:  # Bioware 1 plat
                        da.attrs["description"] = "Data from Bioware 1 plat"
                    else:
                        raise Exception("The number of Force variables is not 3 or 6")
                    da.attrs["freq"] = freq
                    da.attrs["units"] = units
                    # da.time.attrs['units']='s'
                    # da.plot.line(x='time', col='axis', hue='n_var')
                except Exception as err:
                    da = xr.DataArray()
                    raise Exception(f"Not available force data in file, {err}")

            # EMG
            elif section == "EMG":
                labels = [n for n in analogs_values if "EMG" in n]
                units = analog_infos["units"][labels[0]]
                data = []
                data.append([analogs_values[s] for s in labels])
                data = np.concatenate(data)

                coords = {
                    "n_var": labels,
                    "time": np.arange(data.shape[-1]) / freq,
                }
                da = xr.DataArray(
                    data=data,
                    dims=coords.keys(),
                    coords=coords,
                    attrs={"freq": freq, "units": units},
                )

                # Polars version (not much faster)
                """
                dfData = pl.from_dict(analogs_values)
                dfData = dfData.select(pl.col("^EMG.*$"))
                if len(dfData.columns) == 0:
                    raise Exception("No EMG data found in file")

                data = dfData.to_numpy()

                coords = {                    
                    "time": np.arange(data.shape[0]) / freq,
                    "n_var": dfData.columns,
                }
                da = xr.DataArray(
                    data=data,
                    dims=coords.keys(),
                    coords=coords,
                    attrs={"freq": freq, "units": analog_infos['units'][dfData.columns[0]]},
                )      
                """

        btk.btkCloseAcquisition(h)

        da.time.attrs["units"] = "s"
        da.name = section

        # print('Tiempo {0:.3f} s \n'.format(time.perf_counter()-timerSub))

    except Exception as err:
        da = xr.DataArray()
        raise Exception(f"\nATTENTION. Unable to process {file.name}, {err}\n")

    if n_vars_load is not None and "n_var" in da.coords:
        da = da.sel(n_var=n_vars_load)

    return da


def read_vicon_c3d_xr_global(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
) -> xr.DataArray:
    # if section not in ['Trajectories', 'Model Outputs', 'Forces', 'EMG']:

    if section not in ["Trajectories", "Model Outputs", "Forces", "EMG"]:
        raise Exception(
            'Section header not found. try "Trajectories", "Model Outputs", "Forces" or "EMG"'
        )
        return

    try:
        import c3d
    except:
        raise ImportError("Module c3d not installed.\nInstall with pip install c3d")

    timer = time.time()  # inicia el contador de tiempo

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # se asegura de que la extensión es c3d
    file = file.with_suffix(".c3d")

    try:
        timerSub = time.time()  # inicia el contador de tiempo
        print(f"Loading section {section}, file: {file.name}")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with open(file, "rb") as handle:
                reader = c3d.Reader(handle)

                freq = reader.point_rate
                freq_analog = reader.analog_rate

                points = []
                analog = []
                for i, (_, p, a) in enumerate(reader.read_frames()):
                    points.append(p)
                    analog.append(a)
                    # if not i % 10000 and i:
                    #     print("Extracted %d point frames", len(points))

        # Trajectiories and Modeled outputs
        if section in ["Trajectories", "Model Outputs"]:
            labels = [s.replace(" ", "") for s in reader.point_labels]
            data = np.asarray(points)[:, :, :3]

            coords = {
                "time": np.arange(data.shape[0]) / freq,
                "n_var": labels,
                "axis": ["x", "y", "z"],
            }
            da = xr.DataArray(
                data,  # =np.expand_dims(data, axis=0),
                dims=coords.keys(),
                coords=coords,
                name="Trajectories",
                attrs={
                    "freq": freq,
                    "units": "mm",
                },
            ).transpose("n_var", "axis", "time")
            if "Trajectories" in section:
                # Delete unnamed trajectories and modeled outputs
                daTraj = da.sel(
                    n_var=(
                        ~da.n_var.str.startswith("*") & ~da.n_var.str.contains("USERMO")
                    )
                )
            if "Model Outputs" in section:
                daMod = da.sel(n_var=da.n_var.str.contains("USERMO"))

        # Analogs
        elif section in ["Forces", "EMG"]:
            labels_analog = [
                s.split(".")[0].replace(" ", "") for s in reader.analog_labels
            ]
            data_analog = np.concatenate(analog, axis=1)

            # data_analog.shape
            coords = {
                "n_var": labels_analog,
                "time": np.arange(data_analog.shape[1]) / freq_analog,
            }
            da_analog = xr.DataArray(
                data=data_analog,
                dims=coords.keys(),
                coords=coords,
                attrs={"freq": freq_analog},
            )

            # Forces
            if da_analog.n_var.str.contains(
                "Force"
            ).any():  #'Force' in da_analog.n_var:
                daForces = da_analog.sel(
                    n_var="Force"
                )  # .sel(n_var=da_analog.n_var.str.contains('Force'))
                if len(daForces.n_var) == 3:  # 1 platform
                    x = daForces.isel(n_var=0)
                    y = daForces.isel(n_var=1)
                    z = daForces.isel(n_var=2)
                    daForces = (
                        xr.concat([x, y, z], dim="axis")
                        .assign_coords(n_var="plat1")
                        .assign_coords(axis=["x", "y", "z"])
                        .expand_dims({"n_var": 1})
                    )
                elif len(daForces.n_var) == 6:  # 2 platforms
                    x = daForces.isel(n_var=[0, 3])
                    y = daForces.isel(n_var=[1, 4])
                    z = daForces.isel(n_var=[2, 5])
                    daForces = (
                        xr.concat([x, y, z], dim="axis")
                        .assign_coords(n_var=["plat1", "plat2"])
                        .assign_coords(axis=["x", "y", "z"])
                    )
                    daForces.time.attrs["units"] = "s"
                # da.plot.line(x='time', col='axis', hue='n_var')
            else:
                daFor = xr.DataArray()

            # EMG
            if da_analog.n_var.str.contains("EMG").any():
                daEMG = da_analog.sel(n_var=da_analog.n_var.str.contains("EMG"))
                daEMG.time.attrs["units"] = "s"
                # daEMG.n_var.sortby('n_var')
                # daEMG.plot.line(x='time', col='n_var', col_wrap=3)
            else:
                daEMG = xr.DataArray()

        # da.time.attrs['units']='s'

        print("Tiempo {0:.3f} s \n".format(time.time() - timerSub))

    except Exception as err:
        print("\nATTENTION. Unable to process " + file.name, err, "\n")

    if n_vars_load:
        da = da.sel(n_var=n_vars_load)

    daRet = []
    if "Trajectories" in section:
        daRet.append(daTraj)
    if "Model Outputs" in section:
        daRet.append(
            daMod,
        )
    if "Forces" in section:
        daRet.append(daForces)
    if "EMG" in section:
        daRet.append(daEMG)

    if len(daRet) == 1:
        daRet = daRet[0]

    return daRet  # daTrajec, daModels, daForce, daEMG


def read_vicon_c3d_xr_global_ds(
    file: str | Path,
    section: str | None = None,
    n_vars_load: List[str] | None = None,
    coincidence: str = "similar",
):
    try:
        import c3d
    except:
        raise ImportError("Module c3d not installed.\nInstall with pip install c3d")

    timer = time.time()  # inicia el contador de tiempo

    if isinstance(file, str):
        file = Path(file)

    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    # se asegura de que la extensión es c3d
    file = file.with_suffix(".c3d")

    try:
        timerSub = time.time()  # inicia el contador de tiempo
        print("Loading file: {0:s}".format(file.name))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with open(file, "rb") as handle:
                reader = c3d.Reader(handle)

                freq = reader.point_rate
                freq_analog = reader.analog_rate

                points = []
                analog = []
                for i, (_, p, a) in enumerate(reader.read_frames()):
                    points.append(p)
                    analog.append(a)
                    # if not i % 10000 and i:
                    #     print("Extracted %d point frames", len(points))

                labels = [s.replace(" ", "") for s in reader.point_labels]
                labels_analog = [
                    s.split(".")[0].replace(" ", "") for s in reader.analog_labels
                ]
        data = np.asarray(points)[:, :, :3]
        data_analog = np.concatenate(analog, axis=1)

        # Trajectiories and Modeled outputs
        coords = {
            "time": np.arange(data.shape[0]) / freq,
            "n_var": labels,
            "axis": ["x", "y", "z"],
        }
        da = xr.DataArray(
            data,  # =np.expand_dims(data, axis=0),
            dims=coords.keys(),
            coords=coords,
            name="Trajectories",
            attrs={
                "freq": freq,
                "units": "mm",
            },
        ).transpose("n_var", "axis", "time")
        da.time.attrs["units"] = "s"

        # if section=='Trajectories':
        #     #Delete unnamed trajectories and modeled outputs
        #     da = da.sel(n_var=(~da.n_var.str.startswith('*') & ~da.n_var.str.contains('USERMO')))
        # elif section=='Model Outputs':
        #     da = da.sel(n_var=da.n_var.str.contains('USERMO'))

        daTrajec = da.sel(
            n_var=(~da.n_var.str.startswith("*") & ~da.n_var.str.contains("USERMO"))
        )

        daModels = da.sel(n_var=da.n_var.str.contains("USERMO"))
        # da.isel(axis=0).plot.line(x='time', hue='n_var')

        # Analogs
        # data_analog.shape
        coords = {
            "n_var": labels_analog,
            "time": np.arange(data_analog.shape[1]) / freq_analog,
        }
        da_analog = xr.DataArray(
            data=data_analog,
            dims=coords.keys(),
            coords=coords,
            attrs={"freq": freq_analog},
        )

        # Forces
        if da_analog.n_var.str.contains("Force").any():  #'Force' in da_analog.n_var:
            daForce = da_analog.sel(
                n_var="Force"
            )  # .sel(n_var=da_analog.n_var.str.contains('Force'))
            if len(daForce.n_var) == 3:  # 1 platform
                x = daForce.isel(n_var=0)
                y = daForce.isel(n_var=1)
                z = daForce.isel(n_var=2)
                daForce = (
                    xr.concat([x, y, z], dim="axis")
                    .assign_coords(n_var="plat1")
                    .assign_coords(axis=["x", "y", "z"])
                    .expand_dims({"n_var": 1})
                )
            elif len(daForce.n_var) == 6:  # 2 platforms
                x = daForce.isel(n_var=[0, 3])
                y = daForce.isel(n_var=[1, 4])
                z = daForce.isel(n_var=[2, 5])
                daForce = (
                    xr.concat([x, y, z], dim="axis")
                    .assign_coords(n_var=["plat1", "plat2"])
                    .assign_coords(axis=["x", "y", "z"])
                )
            # daForce.plot.line(x='time', col='axis', hue='n_var')
        else:
            daForce = xr.DataArray()

        # EMG
        if da_analog.n_var.str.contains("EMG").any():
            daEMG = da_analog.sel(n_var=da_analog.n_var.str.contains("EMG"))
            daEMG.n_var.sortby("n_var")
            # daEMG.plot.line(x='time', col='n_var', col_wrap=3)
        else:
            daEMG = xr.DataArray()

        print("Tiempo {0:.3f} s \n".format(time.time() - timerSub))

    except Exception as err:
        print("\nATTENTION. Unable to process " + file.name, err, "\n")

    if n_vars_load:
        da = da.sel(n_var=n_vars_load)

    daTodo = xr.Dataset(
        {"Trajectories": daTrajec, "Modeled": daModels, "Forces": daForce, "EMG": daEMG}
    )

    return daTodo  # daTrajec, daModels, daForce, daEMG


# =============================================================================
# %% MAIN
# =============================================================================
if __name__ == "__main__":
    # from biomdp.io.read_vicon_c3d import read_vicon_c3d

    work_path = Path(r"src\datasets")
    file = work_path / "vicon_CMJ_kinem_kinet_emg.c3d"
    daData = read_vicon_c3d(file, section="Trajectories")
    daData.plot.line(x="time", col="n_var", col_wrap=3)

    daData = read_vicon_c3d(file, section="Model Outputs")
    daData.plot.line(x="time", col="n_var", col_wrap=3, sharey=False)

    daData = read_vicon_c3d(file, section="Forces")
    daData.plot.line(x="time", col="n_var", col_wrap=3)

    daData = read_vicon_c3d(file, section="EMG")
    daData.plot.line(x="time", col="n_var", col_wrap=3)

    import timeit

    def test_performance():
        result = read_vicon_c3d(file, section="Trajectories", engine="c3d")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=10):.4f} s"
    )

    def test_performance():
        result = read_vicon_c3d(file, section="Trajectories", engine="ezc3d")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=10):.4f} s"
    )

    def test_performance():
        result = read_vicon_c3d(file, section="Trajectories", engine="mbbtk")
        return result

    print(
        f"{timeit.timeit('test_performance()', setup='from __main__ import test_performance', number=10):.4f} s"
    )

    ruta_archivo = Path(r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D")
    file = ruta_archivo / "SaltosS13_SJ_100S_03.c3d"
    daTrajec = read_vicon_c3d(file, section="Trajectories", engine="ezc3d")
    daTrajec.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)

    daTrajec_c3d = read_vicon_c3d(file, section="Trajectories", engine="c3d")
    daTrajec_ezc3d = read_vicon_c3d(file, section="Trajectories", engine="ezc3d")
    daTrajec_mbbtk = read_vicon_c3d(file, section="Trajectories", engine="mbbtk")

    daTrajec_ezc3d == daTrajec_mbbtk
    daTrajec_c3d == daTrajec_mbbtk

    daTrajec = read_vicon_c3d(
        file,
        section="Trajectories",
        n_vars_load=["LPSI", "LASIMID", "RASIMID", "RPSI"],
        engine="ezc3d",
    )
    daTrajec.plot.line(x="time", col="n_var")

    daTrajec = read_vicon_c3d(
        file,
        section="Trajectories",
        n_vars_load=["LPSI", "LASIMID", "RASIMID", "RPSI"],
        engine="mbbtk",
    )
    daTrajec.plot.line(x="time", col="n_var")

    daModels = read_vicon_c3d(file, section="Model Outputs", engine="ezc3d")
    # Mezcla variables modeladas de ángulo con EMG (3 canales por variable)
    daModels.isel(n_var=slice(None)).plot.line(
        x="time", col="n_var", col_wrap=3, sharey=False
    )
    # modelos sacados directamente de exportar a csv
    nom_vars = ",,S13:AngArtAnkle_R,,,S13:AngArtHip_R,,,S13:AngArtKnee_R,,,S13:AngSegMUSLO_R,,,S13:AngSegPELVIS_LR,,,S13:AngSegPIERNA_R,,,S13:AngSegPIE_R,,,S13:EMG1,S13:EMG2,S13:EMG3,S13:EMG4,S13:EMG5,S13:EMG6,S13:Forces,,,S13:LASI,,,S13:LHJC,,,S13:RAJC,,,S13:RASI,,,S13:RHJC,,,S13:RKJC,,,S13:Right_AnkleExt,,,S13:Right_AnkleInt,,,S13:Right_KneeExt,,,S13:Right_KneeInt,,,S13:LASI,,,S13:LHJC,,,S13:RAJC,,,S13:RASI,,,S13:RHJC,,,S13:RKJC,,,S13:Right_AnkleExt,,,S13:Right_AnkleInt,,,S13:Right_KneeExt,,,S13:Right_KneeInt,,,".split(
        ","
    )[2::3]
    nom_vars = [s.split(":")[-1] for s in nom_vars]
    # Hay que hacer el ajuste de los nombres de las variables a mano para que coincidan
    replace_vars = dict(zip(daModels.n_var.data[:7], nom_vars[:7]))
    names = pd.DataFrame(daModels.n_var.data).replace(replace_vars)[0].tolist()
    daModels = daModels.assign_coords(n_var=names)

    daModel_forces = daModels.isel(n_var=-1)
    daModel_forces.plot.line(x="time")
    daModel_EMG = daModels.sel(n_var=daModels.n_var.str.contains("USER")).isel(
        n_var=slice(None, -1)
    )
    daModel_EMG.plot.line(x="time", col="n_var", col_wrap=3)

    daForce = read_vicon_c3d(file, section="Forces", engine="ezc3d")
    daForce.plot.line(x="time", col="n_var", col_wrap=3)

    daForce2 = read_vicon_c3d(file, section="Forces", engine="mbbtk")
    daForce2.plot.line(x="time", col="n_var", col_wrap=3)
    daForce.equals(daForce2)

    daEMG = read_vicon_c3d(
        file,
        section="EMG",
        n_vars_load=["EMG1", "EMG2", "EMG3", "EMG4", "EMG5", "EMG6"],
        engine="c3d",
    )
    daEMG = read_vicon_c3d(
        file,
        section="EMG",
        n_vars_load=["EMG1.v", "EMG2.v", "EMG3.v", "EMG4.v", "EMG5.v", "EMG6.v"],
        engine="ezc3d",
    )
    daEMG.plot.line(x="time", col="n_var", col_wrap=3)

    daEMG2 = read_vicon_c3d(
        file,
        section="EMG",
        n_vars_load=["EMG1.v", "EMG2.v", "EMG3.v", "EMG4.v", "EMG5.v", "EMG6.v"],
        engine="mbbtk",
    )
    daEMG2.plot.line(x="time", col="n_var", col_wrap=3)
    daEMG.equals(daEMG2)

    [daTrajec, daModels] = read_vicon_c3d_xr_global(
        file, section=["Trajectories", "Model Outputs"]
    )
    daTrajec.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)
    daModels.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)

    n_vars_load = {
        "Trajectories": ["LPSI", "LASIMID", "RASIMID", "RPSI"],
        "Model Outputs": [
            "USERMO",
            "USERM1",
            "USERM2",
        ],
    }
    [daTrajec, daModels] = read_vicon_c3d_xr_global(
        file, section=["Trajectories", "Model Outputs"]
    )
    daTrajec.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)
    daModels.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)

    ruta_archivo = Path(r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D")
    file = ruta_archivo / "PCFutbol-1poa.c3d"
    daForce = read_vicon_c3d(file, section="Forces")
    daForce.plot.line(x="time", col="n_var", col_wrap=3)
    daForce = read_vicon_c3d(file, section="Forces", engine="mbbtk")
    daForce.plot.line(x="time", col="n_var", col_wrap=3)
    daForce.equals(daForce2)

    daEMG = read_vicon_c3d(
        file,
        section="EMG",
        n_vars_load=["EMG1", "EMG2", "EMG3", "EMG4", "EMG5", "EMG6"],
    )

    ruta_archivo = Path(r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D")
    file = ruta_archivo / "Pablo_FIN.c3d"
    daModels = read_vicon_c3d(file, section="Model Outputs")
    daModels2 = read_vicon_c3d(file, section="Model Outputs", engine="mbbtk")
    daTraj = read_vicon_c3d(file, section="Trajectories", engine="mbbtk")
    daModels.equals(daModels2)
    daModels.isel(n_var=slice(6)).plot.line(x="time", col="n_var", col_wrap=3)
    daForce = read_vicon_c3d(file, section="Forces")
    daForce.plot.line(x="time", col="n_var", col_wrap=3)

    daEMG = read_vicon_c3d(
        file,
        section="EMG",
        n_vars_load=["EMG1", "EMG2", "EMG3", "EMG4", "EMG5", "EMG6"],
    )

    # ---- Compara engines

    ruta_archivo = Path(r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D")
    file = ruta_archivo / "SaltosS13_SJ_100S_03.c3d"

    import timeit

    section = "Trajectories"

    def test_perf():
        result = read_vicon_c3d(file, section=section, engine="c3d")
        return result

    print(
        f"{timeit.timeit('test_perf()', setup='from __main__ import test_perf', number=50):.4f} s"
    )

    def test_perf():
        result = read_vicon_c3d(file, section=section, engine="ezc3d")
        return result

    print(
        f"{timeit.timeit('test_perf()', setup='from __main__ import test_perf', number=50):.4f} s"
    )

    def test_perf():
        result = read_vicon_c3d(file, section=section, engine="mbbtk")
        return result

    print(
        f"{timeit.timeit('test_perf()', setup='from __main__ import test_perf', number=50):.4f} s"
    )



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 02 16:36:37 2023

@author: Jose L. L. Elvira

Bioware's .c3d exports only sensor data separately,
8 per platform.
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.2.1"
__date__ = "20/08/2025"

"""
Updates:
    20/08/2025, v0.2.1
        - Incuded comment to export to c3d properly frequencies from Bioware.

    25/03/2025, v0.2.0
        - Incuded general function to distribute according to "engine".

    11/03/2025, v0.1.1
            - Adapted to biomdp with translations.
    
    10/12/2024, v0.1.0
            - Comprueba si está instalado c3d y ezc3d, si no, avisa con mensaje.
            - Incluida función con ezc3d, pero no carga la serie temporal completa.
    
    29/12/2023, v0.0.2
            - Calcula las fuerzas en los 3 ejes a partir de las raw de los sensores.

    02/10/2023, v0.0.1
            - Empezado a partir de read_vicon.c3d.
            
"""

import time
from pathlib import Path
from typing import List

import numpy as np

# import pandas as pd
import xarray as xr

# =============================================================================
# %% Functions
# =============================================================================


def read_kistler_c3d(
    file: str | Path,
    n_vars_load: List[str] | None = None,
    # coincidence: str = "similar",
    engine: str = "ezc3d",
    raw: bool = False,
) -> xr.DataArray:
    if not file.exists():
        raise FileNotFoundError(f"File {file} not found")

    if engine == "c3d":
        da = read_kistler_c3d_c3d(file)

    elif engine == "ezc3d":
        da = read_kistler_ezc3d(file)

    else:
        print(f"Engine {engine} not implemented. Try 'c3d' or 'ezc3d'")

    if raw:
        if n_vars_load:
            da = da.sel(n_var=n_vars_load)

    else:
        da = split_plataforms(da)
        # da = split_dim_axis(da)
        da = compute_forces_axes(da)
        if n_vars_load:
            da = da.sel(n_var=n_vars_load)

    return da


def read_kistler_c3d_c3d(
    file: str | Path,
) -> xr.DataArray:
    try:
        import c3d
    except:
        raise ImportError("Module c3d not installed.\nInstall with pip install c3d")
    import warnings  # to remove 'no points found' warnings

    # timer = time.perf_counter()

    # se asegura de que la extensión es c3d
    file = file.with_suffix(".c3d")

    try:
        # timerSub = time.perf_counter()
        # print(f'Loading file: {file.name}')

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with open(file, "rb") as handle:
                reader = c3d.Reader(handle)

                # freq = reader.point_rate
                freq_analog = reader.analog_rate

                points = []
                analog = []
                for i, (_, p, a) in enumerate(reader.read_frames()):
                    # points.append(p)
                    analog.append(a)
                    if not i % 10000 and i:
                        print(f"Extracted {len(points)} point frames")

        labels_analog = [s.split(".")[0].replace(" ", "") for s in reader.analog_labels]
        data_analog = np.concatenate(analog, axis=1)
        # import matplotlib.pyplot as plt
        # plt.plot(data_analog.T)
        # data_analog.shape
        coords = {
            "n_var": labels_analog,
            "time": np.arange(data_analog.shape[1]) / freq_analog,
        }
        da_analog = xr.DataArray(
            data=data_analog,
            dims=coords.keys(),
            coords=coords,
            attrs={"freq": freq_analog},
        )

        da_analog.attrs["units"] = "N"
        da_analog.time.attrs["units"] = "s"

        # print(f"Time {time.perf_counter() - timerSub:.3f} s \n")

    except Exception as err:
        print(f"\nATTENTION. Unable to process {file.name}, {err}, \n")

    return da_analog


def read_kistler_ezc3d(
    file: str | Path,
):
    """
    IMPORTANT when exporting from Bioware to c3d, select:
                    Base video frame rate: 1000
                    Measurements per frame: 1
    otherwise reading is not complete
    """
    try:
        import ezc3d
    except:
        raise ImportError(
            "Module ezc3d not installed.\nInstall with pip install ezc3d or conda install -c conda-forge ezc3d"
        )

    # timer = time.perf_counter()

    # se asegura de que la extensión es c3d
    file = file.with_suffix(".c3d")

    try:
        # timerSub = time.perf_counter()
        # print(f'Loading file: {file.name}')

        acq = ezc3d.c3d(file.as_posix())  # , extract_forceplat_data=True)

        freq = acq["parameters"]["ANALOG"]["RATE"]["value"][0]
        labels = acq["parameters"]["ANALOG"]["LABELS"]["value"]
        unit_force = acq["parameters"]["ANALOG"]["UNITS"]["value"][0]
        num_plats = acq["parameters"]["ANALOG"]["USED"]["type"]

        if False:
            acq["parameters"]["FORCE_PLATFORM"]["CORNERS"]["value"]

            plats = acq["data"]["platform"]
            num_plats = len(plats)
            unit_force = plats[0]["unit_force"]

            plat1 = plats[0]["force"][:, :]
            plat1.shape
            plt.plot(data.T)

        data = acq["data"]["analogs"][0]  # analogs
        coords = {
            "n_var": labels,
            "time": np.arange(data.shape[-1]) / freq,
        }
        da = xr.DataArray(
            data=data,
            dims=coords.keys(),
            coords=coords,
            attrs={"freq": freq},
        )

        da.attrs["units"] = unit_force
        da.time.attrs["units"] = "s"

        # print(f"Time {time.perf_counter() - timerSub:.3f} s \n")

    except Exception as err:
        print(f"\nATTENTION. Unable to process {file.name}, {err}\n")

    # if raw:
    #     if n_vars_load:
    #         da = da.sel(n_var=n_vars_load)
    #     return da
    # else:
    #     da = split_plataforms(da)
    #     # da = split_dim_axis(da)
    #     da = compute_forces_axes(da)
    #     if n_vars_load:
    #         da = da.sel(n_var=n_vars_load)

    # if n_vars_load:
    #     da = da.sel(n_var=n_vars_load)

    return da


def split_plataforms(da: xr.DataArray) -> xr.DataArray:
    if len(da.n_var) == 8:  # 1 plate
        da = da.expand_dims("plate").assign_coords(plate=[1])

    elif len(da.n_var) == 16:
        plat1 = da.sel(n_var=da.n_var.str.startswith("F1"))
        plat1 = plat1.assign_coords(n_var=plat1.n_var.str.lstrip("F1"))

        plat2 = da.sel(n_var=da.n_var.str.startswith("F2"))
        plat2 = plat2.assign_coords(n_var=plat2.n_var.str.lstrip("F2"))

        da = xr.concat([plat1, plat2], dim="plate").assign_coords(plate=[1, 2])

    # if plat2.n_var.size == 0:
    #     print("There is only one platform.")
    #     return da.sel(plate=[1])
    # else:
    return da


def split_dim_axis(da: xr.DataArray) -> xr.DataArray:
    # NOT NECESSARY WITH COMPUTE_FORCES_AXES???
    # TODO: The letter of the axis in the name must be removed.
    x = da.sel(n_var=da.n_var.str.contains("x"))
    y = da.sel(n_var=da.n_var.str.contains("y"))
    z = da.sel(n_var=da.n_var.str.contains("z"))
    da = (
        xr.concat([x, y, z], dim="axis")
        # .assign_coords(n_var='plat1')
        .assign_coords(axis=["x", "y", "z"])
        # .expand_dims({'n_var':1})
    )
    return da


def compute_forces_axes(da: xr.DataArray) -> xr.DataArray:
    # da=daForce

    if "plate" not in da.coords:
        da = split_plataforms(da)

    Fx = da.sel(n_var=da.n_var.str.contains("x")).sum(dim="n_var")
    Fy = da.sel(n_var=da.n_var.str.contains("y")).sum(dim="n_var")
    Fz = da.sel(n_var=da.n_var.str.contains("z")).sum(dim="n_var")

    daReturn = xr.concat([Fx, Fy, Fz], dim="axis").assign_coords(axis=["x", "y", "z"])
    # daReturn.plot.line(x='time', col='plate')

    return daReturn


def compute_moments_axes(da: xr.DataArray) -> xr.DataArray:
    # da=daForce
    raise Exception("Not implemented yet.")
    """
    if 'plate' not in da.coords:
        da = split_plataforms(da)

    Fx = da.sel(n_var=da.n_var.str.contains('x')).sum(dim='n_var')
    Fy = da.sel(n_var=da.n_var.str.contains('y')).sum(dim='n_var')
    Fz = da.sel(n_var=da.n_var.str.contains('z')).sum(dim='n_var')
        
    daReturn = (xr.concat([Fx, Fy, Fz], dim='axis')
                .assign_coords(axis=['x', 'y', 'z'])
                )
    #daReturn.plot.line(x='time', col='plate')
    """
    return daReturn


# =============================================================================
# %% TESTS
# =============================================================================
if __name__ == "__main__":
    # from biomdp.io.read_kistler_c3d import read_kistler_c3d_xr

    work_path = Path(r"src\datasets")

    file = work_path / "kistler_CMJ_1plate.c3d"
    daForce = read_kistler_c3d(file, engine="c3d")
    daForce.isel(plate=0).plot.line(x="time")  # , col="plate")

    file = work_path / "kistler_DJ_2plates.c3d"
    daForce = read_kistler_c3d(file, engine="c3d")
    daForce.plot.line(x="time", col="plate")

    daForce = read_kistler_c3d(file, engine="c3d", raw=True)
    daForce.plot.line(x="time")
    daForce = split_plataforms(daForce)
    daForce = compute_forces_axes(daForce)
    daForce.plot.line(x="time", col="plate")

    # With ezc3d does not read the entire file??? Exporting in Bioware, select video frame frequency properly to work with ezc3d
    daForce2 = read_kistler_c3d(file, engine="ezc3d", raw=True)
    daForce2.plot.line(x="time")

    # Compare with c3d and ezc3d
    file = work_path / "kistler_DJ_2plates.c3d"
    daForce = read_kistler_c3d(file, engine="c3d")
    daForce.plot.line(x="time", col="plate")

    daForce2 = read_kistler_c3d(file, engine="ezc3d")
    daForce2.plot.line(x="time", col="plate")

    daForce2.equals(daForce.isel(time=slice(None, daForce2.time.size)))

    ####################
    # TESTS: Exporting in Bioware, select video frame frequency properly to work with ezc3d
    file = Path(
        r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D\Bioware_S09_DJ_30_001_rate1-10.c3d"
    )
    daForce = read_kistler_c3d(file, engine="ezc3d")
    daForce.plot.line(x="time", col="plate")

    file = Path(
        r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D\Bioware_S09_DJ_30_001_rate1000-1000.c3d"
    )
    daForce2 = read_kistler_c3d(file, engine="ezc3d")
    daForce2.plot.line(x="time", col="plate")

    file = Path(
        r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D\Bioware_S09_DJ_30_001_rate1-1.c3d"
    )
    daForce3 = read_kistler_c3d(file, engine="ezc3d")
    daForce3.plot.line(x="time", col="plate")

    # This option works ok: Base video frame rate: 1000
    #                       Measurements per frame: 1
    file = Path(
        r"F:\Programacion\Python\Mios\ViconNexus\C3D\ArchivosC3D\Bioware_S09_DJ_30_001_rate100-1.c3d"
    )
    daForce4 = read_kistler_c3d(file, engine="ezc3d")
    daForce4.plot.line(x="time", col="plate")



# %% -*- coding: utf-8 -*-
"""
Created on Tue Mar 04 16:56:08 2025
Functions to create artificial sinusoidal time series.
Mainly based on xarray.

@author: josel
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

__author__ = "Jose L. L. Elvira"
__version__ = "0.1.0"
__date__ = "05/03/2025"

"""
Updates:
    05/03/2025, v0.1.0
        - Initial version with create_time_series_xr function.

"""

from typing import List
import numpy as np

# import pandas as pd
import xarray as xr
from scipy.signal import butter, filtfilt
from pathlib import Path

import matplotlib.pyplot as plt

# import seaborn as sns


def create_time_series_xr(
    rnd_seed: int | None = None,
    num_subj: int = 10,
    fs: float = 100.0,
    IDini: int = 0,
    range_offset: List[float] = [-2.0, -0.5],
    range_amp: List[float] = [1.0, 2.2],
    range_freq: List[float] = [1.8, 2.4],
    range_af: List[float] = [0.0, 1.0],
    range_duration: List[float] = [5.0, 5.1],
    amplific_noise: List[float] = [0.4, 0.7],
    fc_noise: List[float] = [7.0, 12.0],
) -> xr.DataArray:
    """
    Create a dummy data sample based on sine waves with noise

    Parameters
    ----------
    rnd_seed : int, optional
        Seed for random number generation. Default is None.
    num_subj : int, optional
        Number of subjects. Default is 10.
    fs : float, optional
        Sampling frequency. Default is 100.0.
    IDini : int, optional
        Initial ID number. Default is 0.
    range_offset : list of float, optional
        Range of offset values. Default is [-2.0, -0.5].
    range_amp : list of float, optional
        Range of amplitude values. Default is [1.0, 2.2].
    range_freq : list of float, optional
        Range of frequency values. Default is [1.8, 2.4].
    range_af : list of float, optional
        Range of phase angle values in degrees. Default is [0.0, 1.0].
    range_duration : list of float, optional
        Range of duration values. Default is [5.0, 5.1].
    amplific_noise : list of float, optional
        Range of amplitude values for the noise. Default is [0.4, 0.7].
    fc_noise : list of float, optional
        Range of cut-off frequency values for the noise. Default is [7.0, 12.0].

    Returns
    -------
    xarray.DataArray
        Data array of shape (num_subj, time) with the dummy data.
    """

    if rnd_seed is not None:
        # To maintain consistency when creating the random data
        np.random.seed(rnd_seed)
    subjects = []
    for subj in range(num_subj):
        # print(subj)
        a = np.random.uniform(range_amp[0], range_amp[1])
        of = np.random.uniform(range_offset[0], range_offset[1])
        f = np.random.uniform(range_freq[0], range_freq[1])
        af = np.deg2rad(
            np.random.uniform(range_af[0], range_af[1])
        )  # transform to radians
        err = a * np.random.uniform(amplific_noise[0], amplific_noise[1])
        fc_err = np.random.uniform(fc_noise[0], fc_noise[1])
        duration = np.random.uniform(range_duration[0], range_duration[1])

        Ts = 1.0 / fs  # time interval between data in seconds
        t = np.arange(0, duration, Ts)

        senal = np.array(of + a * np.sin(2 * np.pi * f * t + af))

        # Create a controlled random noise
        pasadas = 2.0  # nº de pasadas del filtro adelante y atrás
        orden = 2
        Cf = (2 ** (1 / pasadas) - 1) ** (
            1 / (2 * orden)
        )  # correqtion factor. Para 2nd order = 0.802
        Wn = 2 * fc_err / fs / Cf
        b1, a1 = butter(orden, Wn, btype="low")
        noise = filtfilt(b1, a1, np.random.uniform(a - err, a + err, len(t)))

        #################################
        subjects.append(senal + noise)
        # subjects.append(np.expand_dims(senal + noise, axis=0))
        # sujeto.append(pd.DataFrame(senal + noise, columns=['value']).assign(**{'ID':'{0:02d}'.format(subj+IDini), 'time':np.arange(0, len(senal)/fs, 1/fs)}))

    # Pad data to last the same
    from itertools import zip_longest

    data = np.array(list(zip_longest(*subjects, fillvalue=np.nan)))

    data = xr.DataArray(
        data=data,
        coords={
            "time": np.arange(data.shape[0]) / fs,
            "ID": [
                f"{i:0>2}" for i in range(IDini, IDini + num_subj)
            ],  # pad with zeros on the left.
        },
    )
    return data


# =============================================================================
# %% TESTS
# =============================================================================

if __name__ == "__main__":

    # ----Create single sample
    n = 6  # number of subjects
    duration = 5  # duration in seconds
    freq = 200.0  # sampling frequency

    data = create_time_series_xr(
        num_subj=n,
        fs=freq,
        IDini=1,
        range_offset=[25, 29],
        range_amp=[40, 45],
        range_freq=[1.48, 1.52],
        range_af=[0, 30],
        amplific_noise=[0.4, 0.7],
        fc_noise=[3.0, 3.5],
        range_duration=[duration, duration],
    )
    data.plot.line(x="time")
    plt.show()

    # ----Create multiple samples with pre and post
    rnd_seed = np.random.seed(12340)
    n = 5
    duration = 5

    freq = 200.0
    pre_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[25, 29],
            range_amp=[40, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )

    post_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[22, 26],
            range_amp=[36, 40],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )

    var_a = xr.concat([pre_a, post_a], dim="moment")

    var_a.sel(n_var="a").plot.line(x="time", col="moment")
    plt.show()

    # ----Create multiple samples with pre and post and multiple variables
    rnd_seed = np.random.seed(12340)
    n = 5
    duration = 5

    freq = 200.0
    pre_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[25, 29],
            range_amp=[40, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    post_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[22, 26],
            range_amp=[36, 40],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_a = xr.concat([pre_a, post_a], dim="moment")

    pre_b = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[35, 39],
            range_amp=[50, 55],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["b"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    post_b = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[32, 36],
            range_amp=[32, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["b"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_b = xr.concat([pre_b, post_b], dim="moment")

    pre_c = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[35, 39],
            range_amp=[10, 15],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["c"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    post_c = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[32, 36],
            range_amp=[12, 16],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["c"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_c = xr.concat([pre_c, post_c], dim="moment")

    # Concat all the subjects
    daTotal = xr.concat([var_a, var_b, var_c], dim="n_var")
    daTotal.name = "Angle"
    daTotal.attrs["freq"] = 1 / (daTotal.time[1].values - daTotal.time[0].values)
    daTotal.attrs["units"] = "deg"
    daTotal.time.attrs["units"] = "s"

    # Plot
    daTotal.plot.line(x="time", col="moment", hue="ID", row="n_var")



# %% -*- coding: utf-8 -*-
"""
Created on Tue Oct 01 13:42:08 2024
Functions to interpolate higher frequencies based on
Shannon reconstruction.

@author: Jose L. L. Elvira
"""


# =============================================================================
# %% LOAD LIBRARIES
# =============================================================================

import numpy as np
import xarray as xr

import matplotlib.pyplot as plt

# from biomdp.general_processing_functions import detrend_dim


__author__ = "Jose L. L. Elvira"
__version__ = "0.1.1"
__date__ = "11/03/2025"

"""
Updates:
    11/03/2025, v0.1.1
        - Adapted to biomdp with translations.

    01/10/2024, v0.1.0
        - Versión inicial basada en la versión C++.

"""

# =============================================================================
# %% Function for Shannon reconstruction
# =============================================================================


def _circularity(data: np.ndarray, overlap: bool = False) -> np.array:
    """
    Create circularity in a single dimension array
    """
    if np.count_nonzero(~np.isnan(data)) == 0:
        return data

    data_recons = np.full(len(data) * 2, np.nan)

    # delete nans
    data = np.array(data[~np.isnan(data)])
    # plt.plot(data)

    if overlap:
        pass

    else:  # without overlapping data
        mitad = int(len(data) / 2)
        if True:  # len(data) % 2 == 0:  # if even
            # Load the first half in reverse order at the beginning
            data_recons[:mitad] = 2 * data[0] - data[mitad:0:-1]

            # Continue loading all data in sequence
            data_recons[mitad : mitad + len(data)] = data

            # Load the second half in reverse order at the end
            data_recons[mitad + len(data) - 1 :] = 2 * data[-1] - data[: mitad - 2 : -1]

            """
            plt.plot(data_recons)

            import pandas as pd

            pd.DataFrame([2 * data[0] - data[mitad:0:-1], data, 2 * data[-1] - data[: mitad - 2 : -1]]).T.plot()
            pd.concat(
                [
                    pd.DataFrame([data[mitad:0:-1]]).T,
                    pd.DataFrame([data]).T,
                    pd.DataFrame([data[-1:mitad:-1]]).T,
                ]
            ).plot()
            """

        else:  # if odd. TODO: Check if it is necessary
            par_impar = 1
            # Load the first half in reverse order at the beginning
            data_recons[: mitad + par_impar] = 2 * data[0] - data[mitad:par_impar:-1]

            # Continue loading all data in sequence
            data_recons[mitad : mitad + len(data)] = data
            data[-1]
            # Load the second half in reverse order at the end
            data_recons[mitad + len(data) - 1 :] = 2 * data[-1] - data[: mitad - 2 : -1]

    """
    else
            {
                DatCircular = new double[DatLeido.Length * 2];     //array con los datos cortados en circularidad sin línea tendencia.
                DatTratado = new double[DatLeido.Length * 2];  //array con la corrección de la línea de tendencia para que coincida inicio y final. Sea par o impar, el nº final es el doble -2.

                //primero comprueba si es par o impar
                int resto;
                Math.DivRem(DatLeido.Length, 2, out resto);

                if (resto == 0) //nº de datos PAR
                {
                    //carga la 1ª mitad en orden inverso al principio                    
                    for (int i = 0; i < (int)(DatLeido.Length / 2); i++)
                    {
                        DatCircular[i] = 2 * DatLeido[0] - DatLeido[(int)(DatLeido.Length / 2) - i - 1];//DatLeido[GNumVarGrafY][0] - DatLeido[GNumVarGrafY][i] + DatLeido[GNumVarGrafY][0];
                    }
                    //carga todos los datos seguidos
                    for (int i = 0; i < (int)(DatLeido.Length) - 1; i++)
                    {
                        DatCircular[(int)(DatLeido.Length / 2) + i] = DatLeido[i];
                    }
                    //carga la 2ª mitad en orden inverso al final
                    for (int i = 0; i < (int)(DatLeido.Length / 2) ; i++)
                    {
                        DatCircular[(int)(DatCircular.Length) - i - 1] = 2 * DatLeido[DatLeido.Length - 1] - DatLeido[i + (int)(DatLeido.Length / 2)];//DatLeido[GNumVarGrafY][DatLeido[GNumVarGrafY].Length - 1] - DatLeido[GNumVarGrafY][i] + DatLeido[GNumVarGrafY][DatLeido[GNumVarGrafY].Length - 1];
                    }
                    //el último lo mete a mano
                    DatCircular[(int)(DatCircular.Length) - (int)(DatLeido.Length / 2) - 1] = DatLeido[DatLeido.Length - 1];
                }
                else //nº de datos IMPAR
                {
                    //carga la 1ª mitad en orden inverso al principio                    
                    for (int i = 0; i < (int)(DatLeido.Length / 2) + 1; i++)
                    {
                        DatCircular[i] = 2 * DatLeido[0] - DatLeido[(int)(DatLeido.Length / 2) - i];//DatLeido[GNumVarGrafY][0] - DatLeido[GNumVarGrafY][i] + DatLeido[GNumVarGrafY][0];
                    }
                    //carga todos los datos seguidos
                    for (int i = 0; i < (int)(DatLeido.Length); i++)
                    {
                        DatCircular[(int)(DatLeido.Length / 2) + i + 1] = DatLeido[i];
                    }
                    //carga la 2ª mitad en orden inverso al final
                    for (int i = 0; i < (int)(DatLeido.Length / 2) ; i++)
                    {
                        DatCircular[(int)(DatCircular.Length) - i - 1] = 2 * DatLeido[DatLeido.Length - 1] - DatLeido[i + (int)(DatLeido.Length / 2) + 1];//DatLeido[GNumVarGrafY][DatLeido[GNumVarGrafY].Length - 1] - DatLeido[GNumVarGrafY][i] + DatLeido[GNumVarGrafY][DatLeido[GNumVarGrafY].Length - 1];
                    }
                }
                
            }

    """
    return data_recons


def create_circularity_xr(
    daData: xr.DataArray, freq: float = None, overlap: bool = False
) -> xr.DataArray:
    """
    Create circularity in signal
    """

    """
    data = daData[0,0,0].data
    ini = daWindow[0,0,0].sel(event='ini').data
    fin = daWindow[0,0,0].sel(event='fin').data
    """
    # daRecortado = recorta_ventana_analisis(daData, daWindow)
    daCirc = xr.apply_ufunc(
        _circularity,
        daData,
        overlap,
        input_core_dims=[["time"], []],
        output_core_dims=[["newtime"]],
        vectorize=True,
    ).rename({"newtime": "time"})

    freq = daData.freq if freq is None else freq
    daCirc = daCirc.assign_coords(time=np.arange(0, len(daCirc.time)) / freq)

    return daCirc


def _detrend(data: np.array) -> np.array:
    """
    Detrend the signal along a single dimension
    """

    trend_line = np.linspace(data[0], data[-1], len(data))
    # plt.plot(trend_line)
    # plt.plot(data)
    # plt.plot(data - trend_line)
    return data - trend_line


def detrend_xr(
    daData: xr.DataArray, freq: float | None = None, overlap: bool | None = False
) -> xr.DataArray:
    """
    Create circularity in signal
    """

    """
    data = daData[0,0,0].data
    ini = daWindow[0,0,0].sel(event='ini').data
    fin = daWindow[0,0,0].sel(event='fin').data
    """
    # daRecortado = recorta_ventana_analisis(daData, daWindow)
    data_tratado = xr.apply_ufunc(
        _detrend,
        daData,
        input_core_dims=[["time"]],
        output_core_dims=[["time"]],
        vectorize=True,
    )

    return data_tratado


def _shannon_reconstruction(
    data: np.array,
    old_freq: int | float | None = None,
    new_freq: int | float | None = None,
) -> np.array:
    """
    Reconstruct signal with Shannon reconstruction
    """
    if old_freq is None or new_freq is None:
        raise ValueError("You have to specify both old_freq and new_freq.")

    delta = 1 / old_freq
    new_delta = 1 / new_freq

    tiempo_muestra = (
        len(data) - 1
    ) * delta  # data length is doubled due to circularity
    # num_puntos_nuevo = int(tiempo_muestra / new_delta) + 1
    num_puntos_nuevo = int(len(data) * new_freq / old_freq)

    fc = 1 / (2 * delta)  # Nyquist frequency

    """
    tpo=time.perf_counter()
    for i in range(50):
        # for loop version
        data_tratado = np.zeros(num_puntos_nuevo)
        for i in range(num_puntos_nuevo):
            t = i * new_delta
            for n in range(len(data)):
                if t - n * delta != 0:
                    m = np.sin(2 * np.pi * fc * (t - n * delta)) / (np.pi * (t - n * delta))
                else:
                    m = 1 / delta
                data_tratado[i] += data[n] * m
            data_tratado[i] *= delta
    print(time.perf_counter()-tpo)
    """

    # Vectorized version (~ x150 faster)
    # Create a grid of time points for the new signal
    t = np.arange(num_puntos_nuevo) * new_delta

    # Create a matrix of time differences (t - n*delta)
    n = np.arange(len(data))
    time_diff_matrix = t[:, np.newaxis] - n * delta

    # Calculate the sinc function values, handling the case where t - n*delta = 0
    with np.errstate(divide="ignore", invalid="ignore"):
        m = np.sin(2 * np.pi * fc * time_diff_matrix) / (np.pi * time_diff_matrix)
    m[np.isnan(m)] = 1 / delta  # Replace NaN values with 1/delta

    # Perform the reconstruction using matrix multiplication
    data_tratado = delta * np.dot(m, data)

    """
    {
        double delta = 1 / (double)frecMuestreo;            //tiempo entre datos en la frecuencia de muestreo original
        double newDelta = 1 / (double)nuevaFrecMuestreo;    //tiempo entre datos en la frecuencia de muestreo nueva 
        

        double tiempoMuestra = (double)(DatLeido.Length-1) * delta;  //el datleido que llega está duplicado de tiempo por la circularidad
        int NumPuntosNuevo = (int)((DatLeido.Length/*-1???*/) * nuevaFrecMuestreo / frecMuestreo);// (int)(tiempoMuestra / newDelta)+2;//estaba con +1
        double fc = 1 / (2 * delta);    //Nyquist frecuency
        //double NyquistFrec = (double)(frecMuestreo/2);
        //double NyquistFrec2 = 2 * NyquistFrec;

        double t, m;

        DatTratado = new double[NumPuntosNuevo];

        progressBar2.Maximum = NumPuntosNuevo + 1;
        progressBar2.Value = 0;


        for (int i = 0; i < NumPuntosNuevo; i++)
        {
            t = (double)i * newDelta;
            for (int n = 0; n < DatLeido.Length; n++)
            {
                if (t - (double)n * delta != 0)
                    m = Math.Sin(2 * Math.PI * fc * (t - (double)n * delta)) / (Math.PI * (t - (double)n * delta));
                else
                    m = 1 / delta;

                DatTratado[i] = DatTratado[i] + DatLeido[n] * m;

                //progressBar2.Value++;
            }
            DatTratado[i] = DatTratado[i] * delta;
            progressBar2.Value++;
        }
    }
    """

    return data_tratado


def shannon_reconstruction_xr(
    daData: xr.DataArray,
    old_freq: int | float | None = None,
    new_freq: int | float | None = None,
) -> xr.DataArray:
    """
    Reconstruct with Shannon
    """

    """
    data = daData[0,0,0].data
    ini = daWindow[0,0,0].sel(event='ini').data
    fin = daWindow[0,0,0].sel(event='fin').data
    """
    # daRecortado = recorta_ventana_analisis(daData, daWindow)

    old_freq = daData.freq if old_freq is None else old_freq

    data_tratado = xr.apply_ufunc(
        _shannon_reconstruction,
        daData,
        old_freq,
        new_freq,
        input_core_dims=[["time"], [], []],
        output_core_dims=[["time"]],
        vectorize=True,
    )

    return data_tratado


# =============================================================================
# %% TESTS
# =============================================================================

if __name__ == "__main__":
    # =============================================================================
    # %%---- Create a sample
    # =============================================================================

    import numpy as np
    import time

    # import pandas as pd
    import xarray as xr
    from scipy.signal import butter, filtfilt
    from pathlib import Path

    import matplotlib.pyplot as plt

    # import seaborn as sns
    from biomdp.create_time_series import create_time_series_xr

    rnd_seed = np.random.seed(12340)
    n = 5
    duration = 5

    freq = 200.0
    Pre_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[25, 29],
            range_amp=[40, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    Post_a = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[22, 26],
            range_amp=[36, 40],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["a"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_a = xr.concat([Pre_a, Post_a], dim="moment")

    Pre_b = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[35, 39],
            range_amp=[50, 55],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["b"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    Post_b = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[32, 36],
            range_amp=[32, 45],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["b"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_b = xr.concat([Pre_b, Post_b], dim="moment")

    Pre_c = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[35, 39],
            range_amp=[10, 15],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["c"], "moment": ["pre"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    Post_c = (
        create_time_series_xr(
            rnd_seed=rnd_seed,
            num_subj=n,
            fs=freq,
            IDini=0,
            range_offset=[32, 36],
            range_amp=[12, 16],
            range_freq=[1.48, 1.52],
            range_af=[0, 30],
            amplific_noise=[0.4, 0.7],
            fc_noise=[3.0, 3.5],
            range_duration=[duration, duration],
        )
        .expand_dims({"n_var": ["c"], "moment": ["post"]})
        .transpose("ID", "moment", "n_var", "time")
    )
    var_c = xr.concat([Pre_c, Post_c], dim="moment")

    # concatena todos los sujetos
    daAll = xr.concat([var_a, var_b, var_c], dim="n_var")
    daAll.name = "Angle"
    daAll.attrs["freq"] = 1 / (
        daAll.time[1].values - daAll.time[0].values
    )  # incluimos la frecuencia como atributo
    daAll.attrs["units"] = "deg"
    daAll.time.attrs["units"] = "s"

    # Graphs
    daAll.plot.line(x="time", col="moment", hue="ID", row="n_var")

    # =============================================================================
    # %% Test the functions
    # =============================================================================
    """
    data = daAll[0, 0, 0, 120:201].data
    data2 = _circularity(data)

    data2 = _detrend(data2)
    plt.plot(data2)
    data2[0]
    data2[-1]
    """
    daAllSlice = daAll.isel(time=slice(100, 201))
    daAllSlice.plot.line(x="time", col="moment", hue="ID", row="n_var")

    daAllSliceCirc = create_circularity_xr(daAllSlice)
    daAllSliceCirc.plot.line(x="time", col="moment", hue="ID", row="n_var")

    daAllSliceCircDetrend = _detrend(daAllSliceCirc[0, 0, 0, :].data)
    plt.plot(daAllSliceCircDetrend)

    daAllSliceCircDetrend = detrend_xr(daAllSliceCirc)
    daAllSliceCircDetrend.plot.line(x="time", col="moment", hue="ID", row="n_var")

    data = daAllSliceCircDetrend[0, 0, 0, :].data
    plt.plot(data)

    # Misma forma original y reconstnoise
    dataShannon = _shannon_reconstruction(data, old_freq=daAll.freq, new_freq=400.0)
    plt.plot(np.arange(len(dataShannon)) / 400, dataShannon, label="Shannon rec")
    plt.plot(np.arange(len(data)) / daAll.freq, data, label="Original", ls="--")
    plt.legend()
    plt.show()

    dataShannon = _shannon_reconstruction(data, old_freq=daAll.freq, new_freq=800.0)
    plt.plot(
        np.arange(len(dataShannon[-100:])) / 400,
        dataShannon[-100:],
        label="Shannon rec",
        marker="*",
    )
    plt.plot(
        np.arange(len(data[-25:])) / daAll.freq,
        data[-25:],
        label="Original",
        ls="--",
        marker="*",
    )
    plt.legend()
    plt.show()
